"""Here is defined the Leaf class."""

from __future__ import annotations

import json
import math
import warnings
from typing import Any, Literal, NamedTuple, Union, TYPE_CHECKING
from pathlib import Path
from functools import lru_cache

import numpy as np

from .node import Node
from .utils import byteorders, lazyattr, SizeType
from .flavor import check_flavor, internal_flavor, toarray
from .flavor import alias_map as flavor_alias_map
from .filters import Filters
from .exceptions import (
    NoSuchChunkError,
    NotChunkAlignedError,
    NotChunkedError,
    PerformanceWarning,
)

if TYPE_CHECKING:
    from .group import Group


# These should be declared as type aliases,
# but ``TypeAlias`` requires Python >= 3.10
# and ``type`` statements require Python >= 3.12.

# ``np.typing.NDArray[np.uint8]`` requires NumPy >= 1.21.
NPByteArray = np.ndarray[tuple[int], np.dtype[np.uint8]]

# ``Buffer`` requires Python >= 3.12.
BufferLike = Union[bytes, bytearray, memoryview, NPByteArray]


def read_cached_cpu_info() -> dict[str, Any]:
    """Load the CPU information from a cache file."""
    try:
        with open(Path.home() / ".pytables-cpuinfo.json") as f:
            return json.load(f)
    except FileNotFoundError:
        return {}


def write_cached_cpu_info(cpu_info_dict: dict[str, Any]) -> None:
    """Write CPU information to a cache file."""
    with open(Path.home() / ".pytables-cpuinfo.json", "w") as f:
        json.dump(cpu_info_dict, f, indent=4)


@lru_cache(maxsize=1)
def get_cpu_info() -> dict[str, Any]:
    """Return a dictionary containing CPU information."""
    cached_info = read_cached_cpu_info()
    if cached_info:
        return cached_info

    try:
        import cpuinfo
    except ImportError:
        return {}
    cpu_info_dict = cpuinfo.get_cpu_info()
    try:
        write_cached_cpu_info(cpu_info_dict)
    except OSError:
        # cpu info cannot be stored.
        # will need to be recomputed in the next process
        pass
    return cpu_info_dict


def csformula(expected_mb: int) -> int:
    """Return the fitted chunksize for expected_mb."""
    # For a basesize of 8 KB, this will return:
    # 8 KB for datasets <= 1 MB
    # 1 MB for datasets >= 10 TB
    basesize = 8 * 1024  # 8 KB is a good minimum
    return basesize * int(2 ** math.log10(expected_mb))


def limit_es(expected_mb: int) -> int:
    """Protection against creating too small or too large chunks."""
    if expected_mb < 1:  # < 1 MB
        expected_mb = 1
    elif expected_mb > 10**7:  # > 10 TB
        expected_mb = 10**7
    return expected_mb


def calc_chunksize(expected_mb: int) -> int:
    """Compute the optimum HDF5 chunksize for I/O purposes.

    Rational: HDF5 takes the data in bunches of chunksize length to
    write the on disk. A BTree in memory is used to map structures on
    disk. The more chunks that are allocated for a dataset the larger
    the B-tree. Large B-trees take memory and causes file storage
    overhead as well as more disk I/O and higher contention for the meta
    data cache.  You have to balance between memory and I/O overhead
    (small B-trees) and time to access to data (big B-trees).

    The tuning of the chunksize parameter affects the performance and
    the memory consumed. This is based on my own experiments and, as
    always, your mileage may vary.

    """
    expected_mb = limit_es(expected_mb)
    zone = int(math.log10(expected_mb))
    expected_mb = 10**zone
    chunksize = csformula(expected_mb)
    # XXX: Multiply by 8 seems optimal for sequential access
    return chunksize * 8


class ChunkInfo(NamedTuple):
    """Information about storage for a given chunk.

    It may also refer to a chunk which is within the dataset's shape but that
    does not exist in storage, i.e. a missing chunk.

    An instance of this named tuple class contains the following information,
    in field order:

    .. attribute:: start

        The coordinates in dataset items where the chunk starts, a tuple of
        integers with the same rank as the dataset.  These coordinates are
        always aligned with chunk boundaries.  Also present for missing
        chunks.

    .. attribute:: filter_mask

        An integer where each active bit signals that the filter in its
        position in the pipeline was disabled when storing the chunk.  For
        instance, ``0b10`` disables shuffling, ``0b100`` disables szip, and so
        on.  ``None`` for missing chunks.

    .. attribute:: offset

        An integer which indicates the offset in bytes of chunk data as it
        exists in storage.  ``None`` for missing chunks.

    .. attribute:: size

        An integer which indicates the size in bytes of chunk data as it
        exists in storage.  ``None`` for missing chunks.

    """

    start: tuple[int, ...] | None
    filter_mask: int | None
    offset: int | None
    size: int | None


class Leaf(Node):
    """Abstract base class for all PyTables leaves.

    A leaf is a node (see the Node class in :class:`Node`) which hangs from a
    group (see the Group class in :class:`Group`) but, unlike a group, it can
    not have any further children below it (i.e. it is an end node).

    This definition includes all nodes which contain actual data (datasets
    handled by the Table - see :ref:`TableClassDescr`, Array -
    see :ref:`ArrayClassDescr`, CArray - see :ref:`CArrayClassDescr`, EArray -
    see :ref:`EArrayClassDescr`, and VLArray - see :ref:`VLArrayClassDescr`
    classes) and unsupported nodes (the UnImplemented
    class - :ref:`UnImplementedClassDescr`) these classes do in fact inherit
    from Leaf.


    .. rubric:: Leaf attributes

    These instance variables are provided in addition to those in Node
    (see :ref:`NodeClassDescr`):

    .. attribute:: byteorder

        The byte ordering of the leaf data *on disk*.  It will be either
        ``little`` or ``big``.

    .. attribute:: dtype

        The NumPy dtype that most closely matches this leaf type.

    .. attribute:: extdim

        The index of the enlargeable dimension (-1 if none).

    .. attribute:: nrows

        The length of the main dimension of the leaf data.

    .. attribute:: nrowsinbuf

        The number of rows that fit in internal input buffers.

        You can change this to fine-tune the speed or memory
        requirements of your application.

    .. attribute:: shape

        The shape of data in the leaf.

    """

    # These are a little hard to override, but so are properties.
    attrs = Node._v_attrs
    """The associated AttributeSet instance - see :ref:`AttributeSetClassDescr`
    (This is an easier-to-write alias of :attr:`Node._v_attrs`."""
    title = Node._v_title
    """A description for this node
    (This is an easier-to-write alias of :attr:`Node._v_title`)."""

    @property
    def name(self) -> str:
        """Name of the node.

        The name of this node in its parent group (This is an
        easier-to-write alias of :attr:`Node._v_name`).
        """
        return self._v_name

    @property
    def chunkshape(self) -> tuple[int, ...]:
        """HDF5 chunk size for chunked leaves (a tuple).

        This is read-only because you cannot change the chunk size of a
        leaf once it has been created.
        """
        return getattr(self, "_v_chunkshape", None)

    @property
    def object_id(self) -> int:
        """Node identifier, which may change from run to run.

        (This is an easier-to-write alias of :attr:`Node._v_objectid`).

        .. versionchanged:: 3.0
           The *objectID* property has been renamed into *object_id*.

        """
        return self._v_objectid

    @property
    def ndim(self) -> int:
        """Return the number of dimensions of the leaf data.

        .. versionadded: 2.4
        """
        return len(self.shape)

    @lazyattr
    def filters(self) -> Filters:
        """Filter properties for this leaf.

        See Also
        --------
        Filters

        """
        return Filters._from_leaf(self)

    @property
    def track_times(self) -> bool:
        """Return True if the timestamps for the leaf are recorded.

        If the leaf is not a dataset, this will fail with HDF5ExtError.

        The track times dataset creation property does not seem to
        survive closing and reopening as of HDF5 1.8.17.  Currently,
        it may be more accurate to test whether the ctime for the
        dataset is 0:
        track_times = (leaf._get_obj_timestamps().ctime == 0)
        """
        return self._get_obj_track_times()

    @property
    def maindim(self) -> int:
        """Dimension along which iterators work.

        Its value is 0 (i.e. the first dimension) when the dataset is not
        extendable, and self.extdim (where available) for extendable ones.
        """
        if self.extdim < 0:
            return 0  # choose the first dimension
        return self.extdim

    @property
    def flavor(self) -> Literal["numpy", "python"]:
        """Type of the data object read from this leaf.

        It can be any of 'numpy' or 'python'.

        You can (and are encouraged to) use this property to get, set
        and delete the FLAVOR HDF5 attribute of the leaf. When the leaf
        has no such attribute, the default flavor is used.
        """
        return self._flavor

    @flavor.setter
    def flavor(self, flavor: Literal["numpy", "python"]) -> None:
        self._v_file._check_writable()
        check_flavor(flavor)
        self._v_attrs.FLAVOR = self._flavor = flavor  # logs the change

    @flavor.deleter
    def flavor(self) -> None:
        del self._v_attrs.FLAVOR
        self._flavor = internal_flavor

    @property
    def size_on_disk(self) -> int:
        """Size on disk of the object.

        The size of this leaf's data in bytes as it is stored on disk.  If the
        data is compressed, this shows the compressed size.  In the case of
        uncompressed, chunked data, this may be slightly larger than the amount
        of data, due to partially filled chunks.
        """
        return self._get_storage_size()

    def __init__(
        self,
        parentnode: Group,
        name: str,
        new: bool = False,
        filters: Filters | None = None,
        byteorder: Literal["little", "big", None] = None,
        _log: bool = True,
        track_times: bool = True,
    ) -> None:
        self._v_new = new
        """Is this the first time the node has been created?"""
        self.nrowsinbuf: int | None = None
        """
        The number of rows that fits in internal input buffers.

        You can change this to fine-tune the speed or memory
        requirements of your application.
        """
        self._flavor: Literal["numpy", "python", None] = None
        """Private storage for the `flavor` property."""

        if new:
            # Get filter properties from parent group if not given.
            if filters is None:
                filters = parentnode._v_filters
            self.__dict__["filters"] = filters  # bypass the property

            if byteorder not in (None, "little", "big"):
                raise ValueError(
                    "the byteorder can only take 'little' or 'big' values "
                    "and you passed: %s" % byteorder
                )
            self.byteorder = byteorder
            """The byte ordering of the leaf data *on disk*."""

        self._want_track_times = track_times

        # Existing filters need not be read since `filters`
        # is a lazy property that automatically handles their loading.

        super().__init__(parentnode, name, _log)

    def __len__(self) -> int:
        """Return the length of the main dimension of the leaf data.

        Please note that this may raise an OverflowError on 32-bit platforms
        for datasets having more than 2**31-1 rows.  This is a limitation of
        Python that you can work around by using the nrows or shape attributes.

        """
        return self.nrows

    def __str__(self) -> str:
        """Return the string representation of the object.

        The string representation for this object is its pathname in the
        HDF5 object tree plus some additional metainfo.
        """
        filters = []
        if self.filters.fletcher32:
            filters.append("fletcher32")
        if self.filters.complevel:
            if self.filters.shuffle:
                filters.append("shuffle")
            if self.filters.bitshuffle:
                filters.append("bitshuffle")
            filters.append(f"{self.filters.complib}({self.filters.complevel})")
        return (
            f"{self._v_pathname} ({self.__class__.__name__}"
            f"{self.shape}{', '.join(filters)}) {self._v_title!r}"
        )

    def _g_post_init_hook(self) -> None:
        """Code to be run after node creation and before creation logging.

        This method gets or sets the flavor of the leaf.

        """
        super()._g_post_init_hook()
        if self._v_new:  # set flavor of new node
            if self._flavor is None:
                self._flavor = internal_flavor
            else:  # flavor set at creation time, do not log
                if self._v_file.params["PYTABLES_SYS_ATTRS"]:
                    self._v_attrs._g__setattr("FLAVOR", self._flavor)
        else:  # get flavor of existing node (if any)
            if self._v_file.params["PYTABLES_SYS_ATTRS"]:
                flavor = getattr(self._v_attrs, "FLAVOR", internal_flavor)
                self._flavor = flavor_alias_map.get(flavor, flavor)
            else:
                self._flavor = internal_flavor

    def _calc_chunkshape(
        self, expectedrows: int, rowsize: int, itemsize: int
    ) -> tuple[int, ...]:
        """Calculate the shape for the HDF5 chunk."""
        # In case of a scalar shape, return the unit chunksize
        if self.shape == ():
            return (SizeType(1),)

        # Compute the chunksize
        MB = 1024 * 1024  # noqa: N806
        expected_mb = (expectedrows * rowsize) // MB
        chunksize = calc_chunksize(expected_mb)
        complib = self.filters.complib
        if (
            complib is not None
            and complib.startswith("blosc2")
            and self._c_classid in ("TABLE", "CARRAY", "EARRAY")
        ):
            # Blosc2 can introspect into blocks, so we can increase the
            # chunksize for improving HDF5 perf for its internal btree.
            # For the time being, this has been implemented efficiently
            # just for tables, but in the future *Array objects could also
            # be included.
            # Use a decent default value for chunksize
            chunksize *= 16
            # Now, go explore the L3 size and try to find a smarter chunksize
            cpu_info = get_cpu_info()
            if "l3_cache_size" in cpu_info:
                # In general, is a good idea to set the chunksize equal to L3
                l3_cache_size = cpu_info["l3_cache_size"]
                # cpuinfo sometimes returns cache sizes as strings (like,
                # "4096 KB"), so refuse the temptation to guess and use the
                # value only when it is an actual int.
                # Also, sometimes cpuinfo does not return a correct L3 size;
                # so in general, enforcing L3 > L2 is a good sanity check.
                l2_cache_size = cpu_info.get("l2_cache_size", "Not found")
                if (
                    type(l3_cache_size) is int
                    and type(l2_cache_size) is int
                    and l3_cache_size > l2_cache_size
                ):
                    chunksize = l3_cache_size
            # In Blosc2, the chunksize cannot be larger than 2 GB
            # BLOSC2_MAX_BUFFERSIZE
            if chunksize > 2**31 - 32:
                chunksize = 2**31 - 32

        maindim = self.maindim
        # Compute the chunknitems
        chunknitems = chunksize // itemsize
        # Safeguard against itemsizes being extremely large
        if chunknitems == 0:
            chunknitems = 1
        chunkshape = list(self.shape)
        # Check whether trimming the main dimension is enough
        chunkshape[maindim] = 1
        newchunknitems = np.prod(chunkshape, dtype=SizeType)
        if newchunknitems <= chunknitems:
            chunkshape[maindim] = chunknitems // newchunknitems
        else:
            # No, so start trimming other dimensions as well
            for j in range(len(chunkshape)):
                # Check whether trimming this dimension is enough
                chunkshape[j] = 1
                newchunknitems = np.prod(chunkshape, dtype=SizeType)
                if newchunknitems <= chunknitems:
                    chunkshape[j] = chunknitems // newchunknitems
                    break
            else:
                # Ops, we ran out of the loop without a break
                # Set the last dimension to chunknitems
                chunkshape[-1] = chunknitems

        return tuple(SizeType(s) for s in chunkshape)

    def _calc_nrowsinbuf(self) -> int:
        """Calculate the number of rows that fits on a PyTables buffer."""
        params = self._v_file.params
        # Compute the nrowsinbuf
        rowsize = self.rowsize
        buffersize = params["IO_BUFFER_SIZE"]
        if rowsize != 0:
            nrowsinbuf = buffersize // rowsize
        else:
            nrowsinbuf = 1

        # Safeguard against row sizes being extremely large
        if nrowsinbuf == 0:
            nrowsinbuf = 1
            # If rowsize is too large, issue a Performance warning
            maxrowsize = params["BUFFER_TIMES"] * buffersize
            if rowsize > maxrowsize:
                warnings.warn(
                    f"The Leaf ``{self._v_pathname}`` is exceeding the "
                    f"maximum recommended rowsize ({maxrowsize} bytes); "
                    f"be ready to see PyTables asking for *lots* "
                    f"of memory and possibly slow I/O.  "
                    f"You may want to reduce the rowsize by trimming the "
                    f"value of dimensions that are orthogonal (and preferably "
                    f"close) to the *main* dimension of this leave.  "
                    f"Alternatively, in case you have specified a very "
                    f"small/large chunksize, you may want to "
                    f"increase/decrease it.",
                    PerformanceWarning,
                )
        return nrowsinbuf

    # This method is appropriate for calls to __getitem__ methods
    def _process_range(
        self,
        start: int,
        stop: int,
        step: int,
        dim: int | None = None,
        warn_negstep: bool = True,
    ) -> tuple[int, int, int]:
        if dim is None:
            nrows = self.nrows  # self.shape[self.maindim]
        else:
            nrows = self.shape[dim]

        if warn_negstep and step and step < 0:
            raise ValueError("slice step cannot be negative")

        # if start is not None: start = long(start)
        # if stop is not None: stop = long(stop)
        # if step is not None: step = long(step)

        return slice(start, stop, step).indices(int(nrows))

    # This method is appropriate for calls to read() methods
    def _process_range_read(
        self, start: int, stop: int, step: int, warn_negstep: bool = True
    ) -> tuple[int, int, int]:
        nrows = self.nrows
        if start is not None and stop is None and step is None:
            # Protection against start greater than available records
            # nrows == 0 is a special case for empty objects
            if 0 < nrows <= start:
                raise IndexError(
                    "start of range (%s) is greater than "
                    "number of rows (%s)" % (start, nrows)
                )
            step = 1
            if start == -1:  # corner case
                stop = nrows
            else:
                stop = start + 1
        # Finally, get the correct values (over the main dimension)
        start, stop, step = self._process_range(
            start, stop, step, warn_negstep=warn_negstep
        )
        return (start, stop, step)

    def _g_copy(
        self,
        newparent: Group,
        newname: str,
        recursive: bool,
        _log: bool = True,
        **kwargs,
    ) -> Leaf:
        # Compute default arguments.
        start = kwargs.pop("start", None)
        stop = kwargs.pop("stop", None)
        step = kwargs.pop("step", None)
        title = kwargs.pop("title", self._v_title)
        filters = kwargs.pop("filters", self.filters)
        chunkshape = kwargs.pop("chunkshape", self.chunkshape)
        copyuserattrs = kwargs.pop("copyuserattrs", True)
        stats = kwargs.pop("stats", None)
        if chunkshape == "keep":
            chunkshape = self.chunkshape  # Keep the original chunkshape
        elif chunkshape == "auto":
            chunkshape = None  # Will recompute chunkshape

        # Fix arguments with explicit None values for backwards compatibility.
        if title is None:
            title = self._v_title
        if filters is None:
            filters = self.filters

        # Create a copy of the object.
        new_node, bytes_ = self._g_copy_with_stats(
            newparent,
            newname,
            start,
            stop,
            step,
            title,
            filters,
            chunkshape,
            _log,
            **kwargs,
        )

        # Copy user attributes if requested (or the flavor at least).
        if copyuserattrs:
            self._v_attrs._g_copy(new_node._v_attrs, copyclass=True)
        elif "FLAVOR" in self._v_attrs:
            if self._v_file.params["PYTABLES_SYS_ATTRS"]:
                new_node._v_attrs._g__setattr("FLAVOR", self._flavor)
        new_node._flavor = self._flavor  # update cached value

        # Update statistics if needed.
        if stats is not None:
            stats["leaves"] += 1
            stats["bytes"] += bytes_

        return new_node

    def _g_fix_byteorder_data(
        self, data: np.ndarray, dbyteorder: str
    ) -> np.ndarray:
        """Fix the byteorder of data passed in constructors."""
        dbyteorder = byteorders[dbyteorder]
        # If self.byteorder has not been passed as an argument of
        # the constructor, then set it to the same value of data.
        if self.byteorder is None:
            self.byteorder = dbyteorder
        # Do an additional in-place byteswap of data if the in-memory
        # byteorder doesn't match that of the on-disk.  This is the only
        # place that we have to do the conversion manually. In all the
        # other cases, it will be HDF5 the responsible for doing the
        # byteswap properly.
        if dbyteorder in ["little", "big"]:
            if dbyteorder != self.byteorder:
                # if data is not writeable, do a copy first
                if not data.flags.writeable:
                    data = data.copy()
                data.byteswap(True)
        else:
            # Fix the byteorder again, no matter which byteorder have
            # specified the user in the constructor.
            self.byteorder = "irrelevant"
        return data

    def _point_selection(self, key: list | tuple | np.ndarray) -> np.ndarray:
        """Perform a point-wise selection.

        `key` can be any of the following items:

        * A boolean array with the same shape as self. Those positions
          with True values will signal the coordinates to be returned.

        * A numpy array (or list or tuple) with the point coordinates.
          This has to be a two-dimensional array of size len(self.shape)
          by num_elements containing a list of zero-based values
          specifying the coordinates in the dataset of the selected
          elements. The order of the element coordinates in the array
          specifies the order in which the array elements are iterated
          through when I/O is performed. Duplicate coordinate locations
          are not checked for.

        Return the coordinates array.  If this is not possible, raise a
        `TypeError` so that the next selection method can be tried out.

        This is useful for whatever `Leaf` instance implementing a
        point-wise selection.

        """
        input_key = key
        if type(key) in (list, tuple):
            if isinstance(key, tuple) and len(key) > len(self.shape):
                raise IndexError(f"Invalid index or slice: {key!r}")
            # Try to convert key to a numpy array.  If not possible,
            # a TypeError will be issued (to be catched later on).
            try:
                key = toarray(key)
            except ValueError:
                raise TypeError(f"Invalid index or slice: {key!r}")
        elif not isinstance(key, np.ndarray):
            raise TypeError(f"Invalid index or slice: {key!r}")

        # Protection against empty keys
        if len(key) == 0:
            return np.array([], dtype="i8")

        if key.dtype.kind == "b":
            if not key.shape == self.shape:
                raise IndexError(
                    "Boolean indexing array has incompatible shape"
                )
            # Get the True coordinates (64-bit indices!)
            coords = np.asarray(key.nonzero(), dtype="i8")
            coords = np.transpose(coords)
        elif key.dtype.kind == "i" or key.dtype.kind == "u":
            if len(key.shape) > 2:
                raise IndexError(
                    "Coordinate indexing array has incompatible shape"
                )
            elif len(key.shape) == 2:
                if key.shape[0] != len(self.shape):
                    raise IndexError(
                        "Coordinate indexing array has incompatible shape"
                    )
                coords = np.asarray(key, dtype="i8")
                coords = np.transpose(coords)
            else:
                # For 1-dimensional datasets
                coords = np.asarray(key, dtype="i8")

            # handle negative indices
            base = coords if coords.base is None else coords.base
            if base is input_key:
                # never modify the original "key" data
                coords = coords.copy()

            idx = coords < 0
            coords[idx] = (coords + self.shape)[idx]

            # bounds check
            if np.any(coords < 0) or np.any(coords >= self.shape):
                raise IndexError("Index out of bounds")
        else:
            raise TypeError("Only integer coordinates allowed.")
        # We absolutely need a contiguous array
        if not coords.flags.contiguous:
            coords = coords.copy()
        return coords

    def _check_chunked(self) -> None:
        if self.chunkshape is None:
            raise NotChunkedError("The dataset is not chunked")

    def _check_chunk_within(self, coords: tuple[int, ...]) -> None:
        if len(coords) != self.ndim:
            raise ValueError(
                f"Chunk coordinates do not match dataset shape: "
                f"{coords} !~ {self.shape}"
            )
        if any(c < 0 or c >= s for (c, s) in zip(coords, self.shape)):
            raise IndexError(
                f"Chunk coordinates not within dataset shape: "
                f"{coords} <> {self.shape}"
            )

    def _check_chunk_coords(self, coords: tuple[int, ...]) -> None:
        if any(c % cs for (c, cs) in zip(coords, self.chunkshape)):
            raise NotChunkAlignedError(
                f"Coordinates are not multiples of chunk shape: "
                f"{tuple(coords)} !* {self.chunkshape}"
            )

    # Tree manipulation
    def remove(self) -> None:
        """Remove this node from the hierarchy.

        This method has the behavior described
        in :meth:`Node._f_remove`. Please note that there is no recursive flag
        since leaves do not have child nodes.

        """
        self._f_remove(False)

    def rename(self, newname: str) -> None:
        """Rename this node in place.

        This method has the behavior described in :meth:`Node._f_rename()`.

        """
        self._f_rename(newname)

    def move(
        self,
        newparent: Group | None = None,
        newname: str | None = None,
        overwrite: bool = False,
        createparents: bool = False,
    ) -> None:
        """Move or rename this node.

        This method has the behavior described in :meth:`Node._f_move`

        """
        self._f_move(newparent, newname, overwrite, createparents)

    def copy(
        self,
        newparent: Group | None = None,
        newname: str | None = None,
        overwrite: bool = False,
        createparents: bool = False,
        **kwargs,
    ) -> Leaf:
        """Copy this node and return the new one.

        This method has the behavior described in :meth:`Node._f_copy`. Please
        note that there is no recursive flag since leaves do not have child
        nodes.

        .. warning::

            Note that unknown parameters passed to this method will be
            ignored, so may want to double check the spelling of these
            (i.e. if you write them incorrectly, they will most probably
            be ignored).

        Parameters
        ----------
        title
            The new title for the destination. If omitted or None, the original
            title is used.
        filters : Filters
            Specifying this parameter overrides the original filter properties
            in the source node. If specified, it must be an instance of the
            Filters class (see :ref:`FiltersClassDescr`). The default is to
            copy the filter properties from the source node.
        copyuserattrs
            You can prevent the user attributes from being copied by setting
            this parameter to False. The default is to copy them.
        start, stop, step : int
            Specify the range of rows to be copied; the default is to copy all
            the rows.
        stats
            This argument may be used to collect statistics on the copy
            process. When used, it should be a dictionary with keys 'groups',
            'leaves' and 'bytes' having a numeric value. Their values will be
            incremented to reflect the number of groups, leaves and bytes,
            respectively, that have been copied during the operation.
        chunkshape
            The chunkshape of the new leaf.  It supports a couple of special
            values.  A value of keep means that the chunkshape will be the same
            as original leaf (this is the default).  A value of auto means
            that a new shape will be computed automatically in order to ensure
            the best performance when accessing the dataset through the main
            dimension.  Any other value should be an integer or a tuple
            matching the dimensions of the leaf.

        """
        return self._f_copy(
            newparent,
            newname,
            overwrite,
            createparents=createparents,
            **kwargs,
        )

    def truncate(self, size: int) -> None:
        """Truncate the main dimension to be size rows.

        If the main dimension previously was larger than this size, the extra
        data is lost.  If the main dimension previously was shorter, it is
        extended, and the extended part is filled with the default values.

        The truncation operation can only be applied to *enlargeable* datasets,
        else a TypeError will be raised.

        """
        # A non-enlargeable arrays (Array, CArray) cannot be truncated
        if self.extdim < 0:
            raise TypeError("non-enlargeable datasets cannot be truncated")
        self._g_truncate(size)

    def isvisible(self) -> bool:
        """Return True if this node is visible.

        This method has the behavior described in :meth:`Node._f_isvisible()`.

        """
        return self._f_isvisible()

    # Attribute handling
    def get_attr(self, name: str) -> Any:
        """Get a PyTables attribute from this node.

        This method has the behavior described in :meth:`Node._f_getattr`.

        """
        return self._f_getattr(name)

    def set_attr(self, name: str, value: Any) -> None:
        """Set a PyTables attribute for this node.

        This method has the behavior described in :meth:`Node._f_setattr()`.

        """
        self._f_setattr(name, value)

    def del_attr(self, name: str) -> None:
        """Delete a PyTables attribute from this node.

        This method has the behavior described in :meth:`Node_f_delAttr`.

        """
        self._f_delattr(name)

    # Data handling
    def flush(self) -> None:
        """Flush pending data to disk.

        Saves whatever remaining buffered data to disk. It also releases
        I/O buffers, so if you are filling many datasets in the same
        PyTables session, please call flush() extensively so as to help
        PyTables to keep memory requirements low.

        """
        self._g_flush()

    def chunk_info(self, coords: tuple[int, ...]) -> ChunkInfo:
        """Get storage information about the chunk containing the `coords`.

        The coordinates `coords` are a tuple of integers with the same rank as
        the dataset.

        Return a :class:`ChunkInfo` instance with the information.

        The coordinates need not be aligined with chunk boundaries.  This
        means that this method may be used to get the start coordinates of the
        chunk that contains the item at the given coordinates, for use with
        other direct chunking operations (see :attr:`ChunkInfo.start`).

        If the coordinates are within the dataset's shape but there is no such
        chunk in storage (missing chunk), a :class:`ChunkInfo` with a valid
        ``start`` and ``filter_mask = offset = size = None`` is returned.  If
        the coordinates are beyond the shape, :exc:`IndexError` is raised
        (even if the start of the chunk would fall within the shape).

        Calling this method on a non-chunked dataset raises a
        :exc:`NotChunkedError`.

        """
        self._check_chunked()
        self._check_chunk_within(coords)

        coords = np.array(coords, dtype=SizeType)
        filter_mask, offset, size = self._g_chunk_info(coords)

        # Align coordinates to chunk boundary.
        chunkshape = self.chunkshape
        coords //= chunkshape
        coords *= chunkshape
        return ChunkInfo(tuple(coords.tolist()), filter_mask, offset, size)

    def read_chunk(
        self,
        coords: tuple[int, ...],
        out: bytearray | NPByteArray | None = None,
    ) -> bytes | memoryview:
        """Get the raw chunk that starts at the given `coords` from storage.

        The coordinates `coords` are a tuple of integers with the same rank as
        the dataset.  If they are not multiples of its chunkshape,
        :exc:`NotChunkAlignedError` is raised.

        If a buffer-like `out` argument is given, it receives chunk data.  If
        it has insufficient storage for the chunk, :exc:`ValueError` is raised
        (use :meth:`chunk_info()` to get the required capacity).

        The obtained data is supposed to have gone at storage time through
        dataset filters, minus those in the chunk's filter mask (use
        :meth:`chunk_info()` to get it).

        Return the chunk's raw content, either as a `bytes` instance (if `out`
        is ``None``) or as a `memoryview` over the object given as `out`.

        Reading a chunk within the dataset's shape, but not in storage
        (missing chunk) raises a :exc:`NoSuchChunkError`.  If the chunk is
        beyond the shape, :exc:`IndexError` is raised.

        Calling this method on a non-chunked dataset raises a
        :exc:`NotChunkedError`.

        """
        self._check_chunked()
        self._check_chunk_within(coords)
        self._check_chunk_coords(coords)

        if out is not None:
            out = np.ndarray((len(out),), dtype="u1", buffer=out)

        coords = np.array(coords, dtype=SizeType)
        chunk = self._g_read_chunk(coords, out)
        if chunk is None:
            raise NoSuchChunkError(
                f"Can't read missing chunk at coordinates " f"{tuple(coords)}"
            )
        return chunk.tobytes() if out is None else memoryview(out)

    def write_chunk(
        self, coords: tuple[int, ...], data: BufferLike, filter_mask: int = 0
    ) -> None:
        """Write `data` to storage for the chunk starting at the given `coords`.

        The coordinates `coords` are a tuple of integers with the same rank as
        the dataset.  If they are not multiples of its chunkshape,
        :exc:`NotChunkAlignedError` is raised.

        The content of the buffer-like `data` must already have gone through
        dataset filters, minus those in the given `filter_mask` (which is to
        be saved along data; see :attr:`ChunkInfo.filter_mask`).

        Writing a chunk which is already in storage replaces it, otherwise it
        is added to storage as long as it is within the dataset's shape
        (missing chunk).  This means that you may use :meth:`truncate()` to
        grow an enlargeable dataset cheaply (as no chunk data is written),
        then sparsely write selected chunks in arbitrary order.

        If the chunk is beyond the dataset's shape, :exc:`IndexError` is
        raised.

        Calling this method on a non-chunked dataset raises a
        :exc:`NotChunkedError`.

        """
        self._check_chunked()
        self._check_chunk_within(coords)
        self._check_chunk_coords(coords)

        coords = np.array(coords, dtype=SizeType)
        data = np.ndarray((len(data),), dtype="u1", buffer=data)
        self._g_write_chunk(coords, data, filter_mask)

    def _f_close(self, flush: bool = True) -> None:
        """Close this node in the tree.

        This method has the behavior described in :meth:`Node._f_close`.
        Besides that, the optional argument flush tells whether to flush
        pending data to disk or not before closing.

        """
        if not self._v_isopen:
            return  # the node is already closed or not initialized

        # Only do a flush in case the leaf has an IO buffer.  The
        # internal buffers of HDF5 will be flushed afterwards during the
        # self._g_close() call.  Avoiding an unnecessary flush()
        # operation accelerates the closing for the unbuffered leaves.
        if flush and hasattr(self, "_v_iobuf"):
            self.flush()

        # Close the dataset and release resources
        self._g_close()

        # Close myself as a node.
        super()._f_close()

    def close(self, flush: bool = True) -> None:
        """Close this node in the tree.

        This method is completely equivalent to :meth:`Leaf._f_close`.

        """
        self._f_close(flush)



import sys
import zlib
import itertools

import numpy as np

import tables as tb
from tables.tests import common


class ArrayDirectChunkingTestCase(
    common.TempFileMixin, common.PyTablesTestCase
):
    obj = np.arange(25, dtype="uint8")

    def setUp(self):
        super().setUp()
        self.array = self.h5file.create_array("/", "array", self.obj)

    def test_chunk_info(self):
        self.assertRaises(
            tb.NotChunkedError, self.array.chunk_info, (0,) * self.array.ndim
        )

    def test_read_chunk(self):
        self.assertRaises(
            tb.NotChunkedError, self.array.read_chunk, (0,) * self.array.ndim
        )

    def test_read_chunk_out(self):
        arr = np.zeros(self.obj.shape, dtype=self.obj.dtype)
        self.assertRaises(
            tb.NotChunkedError,
            self.array.read_chunk,
            (0,) * self.array.ndim,
            out=memoryview(arr),
        )

    def test_write_chunk(self):
        arr = self.obj // 2
        self.assertRaises(
            tb.NotChunkedError,
            self.array.write_chunk,
            (0,) * self.array.ndim,
            arr,
        )


# For enlargeable and non-enlargeable datasets.
class DirectChunkingTestCase(common.TempFileMixin, common.PyTablesTestCase):
    # Class attributes:
    shape: tuple[int, ...]
    chunkshape: tuple[int, ...]
    shuffle: bool
    obj: np.ndarray

    # Instance attributes:
    array: tb.Leaf  # set by ``setUp()`` and ``_reopen()``
    filters: tb.Filters

    def setUp(self):
        super().setUp()
        self.filters = tb.Filters(
            complib="zlib", complevel=1, shuffle=self.shuffle
        )

    def modified(self, obj):
        # Return altered copy with same dtype and shape.
        raise NotImplementedError

    def iter_chunks(self):
        chunk_ranges = list(
            range(0, s, cs) for (s, cs) in zip(self.shape, self.chunkshape)
        )
        yield from itertools.product(*chunk_ranges)

    def test_chunk_info_aligned(self):
        for chunk_start in self.iter_chunks():
            chunk_info = self.array.chunk_info(chunk_start)
            self.assertEqual(chunk_info.start, chunk_start)
            self.assertIsNotNone(chunk_info.filter_mask)
            self.assertIsNotNone(chunk_info.offset)
            self.assertIsNotNone(chunk_info.size)

    def test_chunk_info_unaligned(self):
        chunk_info_a = self.array.chunk_info((0,) * self.array.ndim)
        chunk_info_u = self.array.chunk_info((1,) * self.array.ndim)
        self.assertIsNotNone(chunk_info_a.start)
        self.assertEqual(chunk_info_a, chunk_info_u)

    def test_chunk_info_aligned_beyond(self):
        beyond = tuple(
            (1 + s // cs) * cs for (s, cs) in zip(self.shape, self.chunkshape)
        )
        self.assertRaises(IndexError, self.array.chunk_info, beyond)

    def test_chunk_info_unaligned_beyond(self):
        beyond = tuple(
            1 + (1 + s // cs) * cs
            for (s, cs) in zip(self.shape, self.chunkshape)
        )
        self.assertRaises(IndexError, self.array.chunk_info, beyond)

    def shuffled(self, bytes_):
        itemsize = self.obj.dtype.itemsize
        return b"".join(bytes_[d::itemsize] for d in range(itemsize))

    def prepare_chunk(self, bytes_, shuffle=None):
        """Prepare chunk bytes by applying shuffle filter (without compression).

        This returns the bytes as they would appear after decompression,
        to allow comparison with decompressed data from HDF5.
        """
        if shuffle is None:
            shuffle = self.shuffle
        return self.shuffled(bytes_) if shuffle else bytes_

    def compress_chunk(self, bytes_, shuffle=None):
        """Prepare and compress chunk bytes for writing to HDF5."""
        assert self.filters.complib == "zlib"
        prepared = self.prepare_chunk(bytes_, shuffle)
        return zlib.compress(prepared, self.filters.complevel)

    def test_read_chunk(self):
        # Extended to fit chunk boundaries.
        ext_obj = np.pad(
            self.obj,
            [(0, s % cs) for (s, cs) in zip(self.shape, self.chunkshape)],
        )
        for chunk_start in self.iter_chunks():
            chunk = self.array.read_chunk(chunk_start)
            self.assertIsInstance(chunk, bytes)
            obj_slice = tuple(
                slice(s, s + cs)
                for (s, cs) in zip(chunk_start, self.chunkshape)
            )
            # Compare decompressed data to avoid zlib implementation differences
            expected_bytes = self.prepare_chunk(ext_obj[obj_slice].tobytes())
            self.assertEqual(zlib.decompress(chunk), expected_bytes)

    def test_read_chunk_out(self):
        # Extended to fit chunk boundaries.
        ext_obj = np.pad(
            self.obj,
            [(0, s % cs) for (s, cs) in zip(self.shape, self.chunkshape)],
        )
        chunk_start = (0,) * self.obj.ndim
        obj_slice = tuple(
            slice(s, s + cs) for (s, cs) in zip(chunk_start, self.chunkshape)
        )
        expected_bytes = self.prepare_chunk(ext_obj[obj_slice].tobytes())

        # First read the chunk to get its actual compressed size
        chunk_reference = self.array.read_chunk(chunk_start)
        chunk_size = len(chunk_reference)

        chunk_out = bytearray(chunk_size - 1)  # too short
        self.assertRaises(
            ValueError, self.array.read_chunk, chunk_start, out=chunk_out
        )

        chunk_out = bytearray(chunk_size)
        chunk = self.array.read_chunk(chunk_start, out=chunk_out)
        self.assertIsInstance(chunk, memoryview)
        # Compare decompressed data to avoid zlib implementation differences
        self.assertEqual(zlib.decompress(bytes(chunk)), expected_bytes)
        self.assertEqual(zlib.decompress(bytes(chunk_out)), expected_bytes)

    def test_read_chunk_unaligned(self):
        self.assertRaises(
            tb.NotChunkAlignedError,
            self.array.read_chunk,
            (1,) * self.array.ndim,
        )

    def test_read_chunk_beyond(self):
        beyond = tuple(
            (1 + s // cs) * cs for (s, cs) in zip(self.shape, self.chunkshape)
        )
        self.assertRaises(IndexError, self.array.read_chunk, beyond)

    def test_write_chunk(self):
        new_obj = self.modified(self.obj)
        # Extended to fit chunk boundaries.
        ext_obj = np.pad(
            new_obj,
            [(0, s % cs) for (s, cs) in zip(self.shape, self.chunkshape)],
        )
        for chunk_start in self.iter_chunks():
            obj_slice = tuple(
                slice(s, s + cs)
                for (s, cs) in zip(chunk_start, self.chunkshape)
            )
            obj_bytes = self.compress_chunk(ext_obj[obj_slice].tobytes())
            self.array.write_chunk(chunk_start, obj_bytes)

        self._reopen()
        self.assertTrue(common.areArraysEqual(self.array[:], new_obj))

    def test_write_chunk_filtermask(self):
        no_shuffle_mask = 0x00000004  # to turn shuffle off

        chunk_start = (0,) * self.obj.ndim
        obj_slice = tuple(
            slice(s, s + cs) for (s, cs) in zip(chunk_start, self.chunkshape)
        )
        new_obj = self.obj.copy()
        new_obj[obj_slice] = self.modified(new_obj[obj_slice])
        obj_bytes = self.compress_chunk(
            new_obj[obj_slice].tobytes(), shuffle=False
        )
        self.array.write_chunk(
            chunk_start, obj_bytes, filter_mask=no_shuffle_mask
        )

        self._reopen()
        arr_obj = self.array[:]  # first chunk is shuffled, fix it
        fixed_bytes = self.shuffled(arr_obj[obj_slice].tobytes())
        fixed_chunk = np.ndarray(
            self.chunkshape, dtype=self.obj.dtype, buffer=fixed_bytes
        )
        arr_obj[obj_slice] = fixed_chunk
        self.assertTrue(common.areArraysEqual(arr_obj, new_obj))

        chunk_info = self.array.chunk_info(chunk_start)
        self.assertEqual(chunk_info.filter_mask, no_shuffle_mask)

    def test_write_chunk_unaligned(self):
        self.assertRaises(
            tb.NotChunkAlignedError,
            self.array.write_chunk,
            (1,) * self.array.ndim,
            b"foobar",
        )

    def test_write_chunk_beyond(self):
        beyond = tuple(
            (1 + s // cs) * cs for (s, cs) in zip(self.shape, self.chunkshape)
        )
        self.assertRaises(
            IndexError, self.array.write_chunk, beyond, b"foobar"
        )


# For enlargeable datasets only.
class XDirectChunkingTestCase(DirectChunkingTestCase):
    def test_chunk_info_miss_extdim(self):
        # Next chunk in the enlargeable dimension.
        assert self.array.extdim == 0
        chunk_start = (
            ((1 + self.shape[0] // self.chunkshape[0]) * self.chunkshape[0]),
            *((0,) * (self.array.ndim - 1)),
        )
        self.assertRaises(IndexError, self.array.chunk_info, chunk_start)

        # Enlarge the array to put the (missing) chunk within the shape.
        self.array.truncate(chunk_start[0] + self.chunkshape[0])
        chunk_info = self.array.chunk_info(chunk_start)
        self.assertIsNone(chunk_info.filter_mask)
        self.assertIsNone(chunk_info.offset)
        self.assertIsNone(chunk_info.size)

    def test_chunk_info_miss_noextdim(self):
        if self.array.ndim < 2:
            raise common.unittest.SkipTest(
                "missing chunk always within enlargeable dimension"
            )

        # Next chunk in the first non-enlargeable dimension.
        assert self.array.extdim != 1
        chunk_start = (
            0,
            ((1 + self.shape[1] // self.chunkshape[1]) * self.chunkshape[1]),
            *((0,) * (self.array.ndim - 2)),
        )
        self.assertRaises(IndexError, self.array.chunk_info, chunk_start)

    def test_read_chunk_miss_extdim(self):
        # Next chunk in the enlargeable dimension.
        assert self.array.extdim == 0
        chunk_start = (
            ((1 + self.shape[0] // self.chunkshape[0]) * self.chunkshape[0]),
            *((0,) * (self.array.ndim - 1)),
        )
        self.assertRaises(IndexError, self.array.read_chunk, chunk_start)

        # Enlarge the array to put the (missing) chunk within the shape.
        self.array.truncate(chunk_start[0] + self.chunkshape[0])
        self.assertRaises(
            tb.NoSuchChunkError, self.array.read_chunk, chunk_start
        )

    def _test_write_chunk_missing(self, shrink_after):
        # Enlarge array by two chunk rows,
        # copy first old chunk in first chunk of new last chunk row.
        assert self.array.extdim == 0
        chunk_start = (
            (
                (1 + self.shape[0] // self.chunkshape[0]) * self.chunkshape[0]
                + self.chunkshape[0]
            ),
            *((0,) * (self.array.ndim - 1)),
        )
        chunk = self.array.read_chunk((0,) * self.array.ndim)
        self.array.truncate(chunk_start[0] + self.chunkshape[0])
        self.array.write_chunk(chunk_start, chunk)
        if shrink_after:
            self.array.truncate(self.shape[0] + 1)
            self.array.truncate(self.shape[0] - 1)

        new_obj = self.obj.copy()
        new_obj.resize(self.array.shape, refcheck=False)
        obj_slice = tuple(
            slice(s, s + cs) for (s, cs) in zip(chunk_start, self.chunkshape)
        )
        if not shrink_after:
            new_obj[obj_slice] = new_obj[
                tuple(slice(0, cs) for cs in self.chunkshape)
            ]

        self._reopen()
        self.assertTrue(common.areArraysEqual(self.array[:], new_obj))

    def test_write_chunk_missing1(self):
        return self._test_write_chunk_missing(shrink_after=False)

    def test_write_chunk_missing2(self):
        return self._test_write_chunk_missing(shrink_after=True)


class CArrayDirectChunkingTestCase(DirectChunkingTestCase):
    shape = (5, 5)
    chunkshape = (2, 2)  # 3 x 3 chunks, incomplete at right/bottom boundaries
    shuffle = True
    obj = np.arange(np.prod(shape), dtype="u2").reshape(shape)

    def setUp(self):
        super().setUp()
        self.array = self.h5file.create_carray(
            "/",
            "carray",
            chunkshape=self.chunkshape,
            obj=self.obj,
            filters=self.filters,
        )

    def _reopen(self):
        super()._reopen()
        self.array = self.h5file.root.carray

    def modified(self, obj):
        return obj * 2


class EArrayDirectChunkingTestCase(XDirectChunkingTestCase):
    shape = (5, 5)  # enlargeable along first dimension
    chunkshape = (2, 2)  # 3 x 3 chunks, incomplete at right/bottom boundaries
    shuffle = True
    obj = np.arange(np.prod(shape), dtype="u2").reshape(shape)

    def setUp(self):
        super().setUp()
        atom = tb.Atom.from_dtype(self.obj.dtype)
        shape = (0, *self.shape[1:])
        self.array = self.h5file.create_earray(
            "/",
            "earray",
            atom,
            shape,
            chunkshape=self.chunkshape,
            filters=self.filters,
        )
        self.array.append(self.obj)

    def _reopen(self):
        super()._reopen()
        self.array = self.h5file.root.earray

    def modified(self, obj):
        return obj * 2


class TableDirectChunkingTestCase(XDirectChunkingTestCase):
    shape = (5,)  # enlargeable along first dimension
    chunkshape = (2,)  # 3 chunks, incomplete at bottom boundary
    shuffle = True
    obj = np.array(
        [(i, float(i)) for i in range(np.prod(shape))], dtype="u4,f4"
    )

    def setUp(self):
        super().setUp()
        desc, _ = tb.descr_from_dtype(self.obj.dtype)
        self.array = self.h5file.create_table(
            "/",
            "table",
            desc,
            chunkshape=self.chunkshape,
            filters=self.filters,
        )
        self.array.append(self.obj)

    def _reopen(self):
        super()._reopen()
        self.array = self.h5file.root.table

    def modified(self, obj):
        flat = obj.copy().reshape((np.prod(obj.shape),))
        fnames = flat.dtype.names
        for i in range(len(flat)):
            for f in fnames:
                flat[i][f] *= 2
        return flat.reshape(obj.shape)


def suite():
    theSuite = common.unittest.TestSuite()
    niter = 1

    for i in range(niter):
        theSuite.addTest(common.make_suite(ArrayDirectChunkingTestCase))
        theSuite.addTest(common.make_suite(CArrayDirectChunkingTestCase))
        theSuite.addTest(common.make_suite(EArrayDirectChunkingTestCase))
        theSuite.addTest(common.make_suite(TableDirectChunkingTestCase))

    return theSuite


if __name__ == "__main__":
    common.parse_argv(sys.argv)
    common.print_versions()
    common.unittest.main(defaultTest="suite")



"""Support for undoing and redoing actions.

Functions:

* undo(file, operation, *args)
* redo(file, operation, *args)
* move_to_shadow(file, path)
* move_from_shadow(file, path)
* attr_to_shadow(file, path, name)
* attr_from_shadow(file, path, name)

Misc variables:

`__docformat__`
    The format of documentation strings in this module.

"""

from typing import Literal, TYPE_CHECKING

from .path import split_path

if TYPE_CHECKING:
    from .file import File

__docformat__ = "reStructuredText"
"""The format of documentation strings in this module."""


def undo(
    file_: "File",
    operation: Literal["ADDATTR", "CREATE", "DELATTR", "MOVE", "REMOVE"],
    *args: str,
) -> None:
    """Undo action."""
    if operation == "CREATE":
        undo_create(file_, args[0])
    elif operation == "REMOVE":
        undo_remove(file_, args[0])
    elif operation == "MOVE":
        undo_move(file_, args[0], args[1])
    elif operation == "ADDATTR":
        undo_add_attr(file_, args[0], args[1])
    elif operation == "DELATTR":
        undo_del_attr(file_, args[0], args[1])
    else:
        raise NotImplementedError(
            "the requested unknown operation %r can "
            "not be undone; please report this to the "
            "authors" % operation
        )


def redo(
    file_: "File",
    operation: Literal["ADDATTR", "CREATE", "DELATTR", "MOVE", "REMOVE"],
    *args: str,
) -> None:
    """Re-do action."""
    if operation == "CREATE":
        redo_create(file_, args[0])
    elif operation == "REMOVE":
        redo_remove(file_, args[0])
    elif operation == "MOVE":
        redo_move(file_, args[0], args[1])
    elif operation == "ADDATTR":
        redo_add_attr(file_, args[0], args[1])
    elif operation == "DELATTR":
        redo_del_attr(file_, args[0], args[1])
    else:
        raise NotImplementedError(
            "the requested unknown operation %r can "
            "not be redone; please report this to the "
            "authors" % operation
        )


def move_to_shadow(file_: "File", path: str) -> None:
    """Move a node to the set of shadowed ones."""
    node = file_._get_node(path)

    shparent, shname = file_._shadow_name()
    node._g_move(shparent, shname)


def move_from_shadow(file_: "File", path: str) -> None:
    """Move a node fro the set of shadowe dones back to foreground."""
    shparent, shname = file_._shadow_name()
    node = shparent._f_get_child(shname)

    pname, name = split_path(path)
    parent = file_._get_node(pname)
    node._g_move(parent, name)


def undo_create(file_: "File", path: str) -> None:
    """Undo create node."""
    move_to_shadow(file_, path)


def redo_create(file_: "File", path: str) -> None:
    """Re-do create node."""
    move_from_shadow(file_, path)


def undo_remove(file_: "File", path: str) -> None:
    """Undo remove node."""
    move_from_shadow(file_, path)


def redo_remove(file_: "File", path: str) -> None:
    """Re-do remove node."""
    move_to_shadow(file_, path)


def undo_move(file_: "File", origpath: str, destpath: str) -> None:
    """Undo move node."""
    origpname, origname = split_path(origpath)

    node = file_._get_node(destpath)
    origparent = file_._get_node(origpname)
    node._g_move(origparent, origname)


def redo_move(file_: "File", origpath: str, destpath: str) -> None:
    """Re-do move node."""
    destpname, destname = split_path(destpath)

    node = file_._get_node(origpath)
    destparent = file_._get_node(destpname)
    node._g_move(destparent, destname)


def attr_to_shadow(file_: "File", path: str, name: str) -> None:
    """Move an attribute to the shadowed attribute set."""
    node = file_._get_node(path)
    attrs = node._v_attrs
    value = getattr(attrs, name)

    shparent, shname = file_._shadow_name()
    shattrs = shparent._v_attrs

    # Set the attribute only if it has not been kept in the shadow.
    # This avoids re-pickling complex attributes on REDO.
    if shname not in shattrs:
        shattrs._g__setattr(shname, value)

    attrs._g__delattr(name)


def attr_from_shadow(file_: "File", path: str, name: str) -> None:
    """Move an attribute from shadowed attribute set to foreground."""
    shparent, shname = file_._shadow_name()
    shattrs = shparent._v_attrs
    value = getattr(shattrs, shname)

    node = file_._get_node(path)
    node._v_attrs._g__setattr(name, value)

    # Keeping the attribute in the shadow allows reusing it on Undo/Redo.
    # shattrs._g__delattr(shname)


def undo_add_attr(file_: "File", path: str, name: str) -> None:
    """Undo add attribute."""
    attr_to_shadow(file_, path, name)


def redo_add_attr(file_: "File", path: str, name: str) -> None:
    """Re-do add attribute."""
    attr_from_shadow(file_, path, name)


def undo_del_attr(file_: "File", path: str, name: str) -> None:
    """Undo delete attribute."""
    attr_from_shadow(file_, path, name)


def redo_del_attr(file_: "File", path: str, name: str) -> None:
    """Re-do delete attribute."""
    attr_to_shadow(file_, path, name)



"""This test unit checks control of dataset timestamps with track_times."""

import sys
import time
import hashlib
from pathlib import Path

import tables as tb
from tables.tests import common

HEXDIGEST = "2aafb84ab739bb4ae61d2939dc010bfd"


class Record(tb.IsDescription):
    var1 = tb.StringCol(itemsize=4)  # 4-character String
    var2 = tb.IntCol()  # integer
    var3 = tb.Int16Col()  # short integer


class TrackTimesMixin:
    def _add_datasets(self, group, j, track_times):
        # Create a table
        table = self.h5file.create_table(
            group,
            f"table{j}",
            Record,
            title=self.title,
            filters=None,
            track_times=track_times,
        )
        # Get the record object associated with the new table
        d = table.row
        # Fill the table
        for i in range(self.nrows):
            d["var1"] = "%04d" % (self.nrows - i)
            d["var2"] = i
            d["var3"] = i * 2
            d.append()  # This injects the Record values
        # Flush the buffer for this table
        table.flush()

        # Create a couple of arrays in each group
        var1List = [x["var1"] for x in table.iterrows()]
        var3List = [x["var3"] for x in table.iterrows()]

        self.h5file.create_array(
            group, f"array{j}", var1List, f"col {j}", track_times=track_times
        )

        # Create CArrays as well
        self.h5file.create_carray(
            group,
            name=f"carray{j}",
            obj=var3List,
            title=f"col {j + 2}",
            track_times=track_times,
        )

        # Create EArrays as well
        ea = self.h5file.create_earray(
            group,
            f"earray{j}",
            tb.StringAtom(itemsize=4),
            (0,),
            f"col {j + 4}",
            track_times=track_times,
        )
        # And fill them with some values
        ea.append(var1List)

        # Finally VLArrays too
        vla = self.h5file.create_vlarray(
            group,
            f"vlarray{j}",
            tb.Int16Atom(),
            f"col {j + 6}",
            track_times=track_times,
        )
        # And fill them with some values
        vla.append(var3List)


class TimestampTestCase(
    TrackTimesMixin, common.TempFileMixin, common.PyTablesTestCase
):
    title = "A title"
    nrows = 10

    def setUp(self):
        super().setUp()
        self.populateFile()

    def populateFile(self):
        group = self.h5file.root
        for j in range(4):
            track_times = bool(j % 2)
            self._add_datasets(group, j, track_times)

    def test00_checkTimestamps(self):
        """Checking retrieval of timestamps"""

        for pattern in (
            "/table{}",
            "/array{}",
            "/carray{}",
            "/earray{}",
            "/vlarray{}",
        ):
            # Verify that:
            # - if track_times was False, ctime is 0
            # - if track_times was True, ctime is not 0,
            #   and has either stayed the same or incremented
            tracked_ctimes = []
            for j in range(4):
                track_times = bool(j % 2)
                node = pattern.format(j)
                obj = self.h5file.get_node(node)
                # Test property retrieval
                self.assertEqual(obj.track_times, track_times)
                timestamps = obj._get_obj_timestamps()
                self.assertEqual(timestamps.atime, 0)
                self.assertEqual(timestamps.mtime, 0)
                self.assertEqual(timestamps.btime, 0)
                if not track_times:
                    self.assertEqual(timestamps.ctime, 0)
                else:
                    self.assertNotEqual(timestamps.ctime, 0)
                    tracked_ctimes.append(timestamps.ctime)
            self.assertGreaterEqual(tracked_ctimes[1], tracked_ctimes[0])


class BitForBitTestCase(
    TrackTimesMixin, common.TempFileMixin, common.PyTablesTestCase
):
    title = "A title"
    nrows = 10

    def repopulateFile(self, track_times):
        self.h5file.close()
        self.h5file = tb.open_file(self.h5fname, mode="w")
        group = self.h5file.root
        self._add_datasets(group, 1, track_times)
        self.h5file.close()

    def test00_checkReproducibility(self):
        """Checking bit-for-bit reproducibility with no track_times"""

        self.repopulateFile(track_times=False)
        hexdigest_wo_track_1 = self._get_digest(self.h5fname)
        self.repopulateFile(track_times=True)
        hexdigest_w_track_1 = self._get_digest(self.h5fname)
        time.sleep(1)
        self.repopulateFile(track_times=True)
        hexdigest_w_track_2 = self._get_digest(self.h5fname)
        self.repopulateFile(track_times=False)
        hexdigest_wo_track_2 = self._get_digest(self.h5fname)
        self.assertEqual(HEXDIGEST, hexdigest_wo_track_1)
        self.assertEqual(hexdigest_wo_track_1, hexdigest_wo_track_2)
        self.assertNotEqual(hexdigest_wo_track_1, hexdigest_w_track_1)
        self.assertNotEqual(hexdigest_w_track_1, hexdigest_w_track_2)

    def _get_digest(self, filename):
        md5 = hashlib.md5()
        for data in Path(filename).read_bytes():
            md5.update(data)

        hexdigest = md5.hexdigest()

        return hexdigest


def suite():
    theSuite = common.unittest.TestSuite()
    niter = 1
    # common.heavy = 1 # Uncomment this only for testing purposes!

    for i in range(niter):
        theSuite.addTest(common.make_suite(TimestampTestCase))
        theSuite.addTest(common.make_suite(BitForBitTestCase))

    return theSuite


if __name__ == "__main__":
    common.parse_argv(sys.argv)
    common.print_versions()
    common.unittest.main(defaultTest="suite")



"""Test module for different kind of links under PyTables."""

import re
import tempfile
from pathlib import Path

import tables as tb
from tables.tests import common


# Test for hard links
class HardLinkTestCase(common.TempFileMixin, common.PyTablesTestCase):

    def setUp(self):
        super().setUp()
        self._createFile()

    def _createFile(self):
        self.h5file.create_array("/", "arr1", [1, 2])
        group1 = self.h5file.create_group("/", "group1")
        arr2 = self.h5file.create_array(group1, "arr2", [1, 2, 3])
        lgroup1 = self.h5file.create_hard_link("/", "lgroup1", "/group1")
        self.assertIsNotNone(lgroup1)
        larr1 = self.h5file.create_hard_link(group1, "larr1", "/arr1")
        self.assertIsNotNone(larr1)
        larr2 = self.h5file.create_hard_link("/", "larr2", arr2)
        self.assertIsNotNone(larr2)

    def test00_create(self):
        """Creating hard links."""

        self._checkEqualityGroup(
            self.h5file.root.group1, self.h5file.root.lgroup1, hardlink=True
        )
        self._checkEqualityLeaf(
            self.h5file.root.arr1, self.h5file.root.group1.larr1, hardlink=True
        )
        self._checkEqualityLeaf(
            self.h5file.root.lgroup1.arr2,
            self.h5file.root.larr2,
            hardlink=True,
        )

    def test01_open(self):
        """Opening a file with hard links."""

        self._reopen()
        self._checkEqualityGroup(
            self.h5file.root.group1, self.h5file.root.lgroup1, hardlink=True
        )
        self._checkEqualityLeaf(
            self.h5file.root.arr1, self.h5file.root.group1.larr1, hardlink=True
        )
        self._checkEqualityLeaf(
            self.h5file.root.lgroup1.arr2,
            self.h5file.root.larr2,
            hardlink=True,
        )

    def test02_removeLeaf(self):
        """Removing a hard link to a Leaf."""

        # First delete the initial link
        self.h5file.root.arr1.remove()
        self.assertNotIn("/arr1", self.h5file)
        # The second link should still be there
        if common.verbose:
            print("Remaining link:", self.h5file.root.group1.larr1)
        self.assertIn("/group1/larr1", self.h5file)
        # Remove the second link
        self.h5file.root.group1.larr1.remove()
        self.assertNotIn("/group1/larr1", self.h5file)

    def test03_removeGroup(self):
        """Removing a hard link to a Group."""

        if common.verbose:
            print("Original object tree:", self.h5file)
        # First delete the initial link
        self.h5file.root.group1._f_remove(force=True)
        self.assertNotIn("/group1", self.h5file)
        # The second link should still be there
        if common.verbose:
            print("Remaining link:", self.h5file.root.lgroup1)
            print("Object tree:", self.h5file)
        self.assertIn("/lgroup1", self.h5file)
        # Remove the second link
        self.h5file.root.lgroup1._g_remove(recursive=True)
        self.assertNotIn("/lgroup1", self.h5file)
        if common.verbose:
            print("Final object tree:", self.h5file)


# Test for soft links
class SoftLinkTestCase(common.TempFileMixin, common.PyTablesTestCase):

    def setUp(self):
        super().setUp()
        self._createFile()

    def _createFile(self):
        self.h5file.create_array("/", "arr1", [1, 2])
        group1 = self.h5file.create_group("/", "group1")
        arr2 = self.h5file.create_array(group1, "arr2", [1, 2, 3])
        lgroup1 = self.h5file.create_soft_link("/", "lgroup1", "/group1")
        self.assertIsNotNone(lgroup1)
        larr1 = self.h5file.create_soft_link(group1, "larr1", "/arr1")
        self.assertIsNotNone(larr1)
        larr2 = self.h5file.create_soft_link("/", "larr2", arr2)
        self.assertIsNotNone(larr2)

    def test00_create(self):
        """Creating soft links."""

        self._checkEqualityGroup(
            self.h5file.root.group1, self.h5file.root.lgroup1()
        )
        self._checkEqualityLeaf(
            self.h5file.root.arr1, self.h5file.root.group1.larr1()
        )
        self._checkEqualityLeaf(
            self.h5file.root.lgroup1().arr2, self.h5file.root.larr2()
        )

    def test01_open(self):
        """Opening a file with soft links."""

        self._reopen()
        self._checkEqualityGroup(
            self.h5file.root.group1, self.h5file.root.lgroup1()
        )
        self._checkEqualityLeaf(
            self.h5file.root.arr1, self.h5file.root.group1.larr1()
        )
        self._checkEqualityLeaf(
            self.h5file.root.lgroup1().arr2, self.h5file.root.larr2()
        )

    def test02_remove(self):
        """Removing a soft link."""

        # First delete the referred link
        self.h5file.root.arr1.remove()
        self.assertNotIn("/arr1", self.h5file)
        # The soft link should still be there (but dangling)
        if common.verbose:
            print("Dangling link:", self.h5file.root.group1.larr1)
        self.assertIn("/group1/larr1", self.h5file)
        # Remove the soft link itself
        self.h5file.root.group1.larr1.remove()
        self.assertNotIn("/group1/larr1", self.h5file)

    def test03_copy(self):
        """Copying a soft link."""

        # Copy the link into another location
        root = self.h5file.root
        lgroup1 = root.lgroup1
        lgroup2 = lgroup1.copy("/", "lgroup2")
        self.assertIn("/lgroup1", self.h5file)
        self.assertIn("/lgroup2", self.h5file)
        self.assertIn("lgroup2", root._v_children)
        self.assertIn("lgroup2", root._v_links)
        if common.verbose:
            print("Copied link:", lgroup2)
        # Remove the first link
        lgroup1.remove()
        self._checkEqualityGroup(
            self.h5file.root.group1, self.h5file.root.lgroup2()
        )

    def test03_overwrite(self):
        """Overwrite a soft link."""

        # Copy the link into another location
        root = self.h5file.root
        lgroup1 = root.lgroup1
        lgroup2 = lgroup1.copy("/", "lgroup2")
        lgroup2 = lgroup1.copy("/", "lgroup2", overwrite=True)
        self.assertIn("/lgroup1", self.h5file)
        self.assertIn("/lgroup2", self.h5file)
        self.assertIn("lgroup2", root._v_children)
        self.assertIn("lgroup2", root._v_links)
        if common.verbose:
            print("Copied link:", lgroup2)
        # Remove the first link
        lgroup1.remove()
        self._checkEqualityGroup(
            self.h5file.root.group1, self.h5file.root.lgroup2()
        )

    def test04_move(self):
        """Moving a soft link."""

        # Move the link into another location
        lgroup1 = self.h5file.root.lgroup1
        group2 = self.h5file.create_group("/", "group2")
        lgroup1.move(group2, "lgroup2")
        lgroup2 = self.h5file.root.group2.lgroup2
        if common.verbose:
            print("Moved link:", lgroup2)
        self.assertNotIn("/lgroup1", self.h5file)
        self.assertIn("/group2/lgroup2", self.h5file)
        self._checkEqualityGroup(
            self.h5file.root.group1, self.h5file.root.group2.lgroup2()
        )

    def test05_rename(self):
        """Renaming a soft link."""

        # Rename the link
        lgroup1 = self.h5file.root.lgroup1
        lgroup1.rename("lgroup2")
        lgroup2 = self.h5file.root.lgroup2
        if common.verbose:
            print("Moved link:", lgroup2)
        self.assertNotIn("/lgroup1", self.h5file)
        self.assertIn("/lgroup2", self.h5file)
        self._checkEqualityGroup(
            self.h5file.root.group1, self.h5file.root.lgroup2()
        )

    def test06a_relative_path(self):
        """Using soft links with relative paths."""

        # Create new group
        self.h5file.create_group("/group1", "group3")
        # ... and relative link
        lgroup3 = self.h5file.create_soft_link("/group1", "lgroup3", "group3")
        if common.verbose:
            print("Relative path link:", lgroup3)
        self.assertIn("/group1/lgroup3", self.h5file)
        self._checkEqualityGroup(
            self.h5file.root.group1.group3, self.h5file.root.group1.lgroup3()
        )

    def test06b_relative_path(self):
        """Using soft links with relative paths (./ version)"""

        # Create new group
        self.h5file.create_group("/group1", "group3")
        # ... and relative link
        lgroup3 = self.h5file.create_soft_link(
            "/group1", "lgroup3", "./group3"
        )
        if common.verbose:
            print("Relative path link:", lgroup3)
        self.assertIn("/group1/lgroup3", self.h5file)
        self._checkEqualityGroup(
            self.h5file.root.group1.group3, self.h5file.root.group1.lgroup3()
        )

    def test07_walkNodes(self):
        """Checking `walk_nodes` with `classname` option."""

        links = [
            node._v_pathname
            for node in self.h5file.walk_nodes("/", classname="Link")
        ]
        if common.verbose:
            print("detected links (classname='Link'):", links)
        self.assertEqual(links, ["/larr2", "/lgroup1", "/group1/larr1"])
        links = [
            node._v_pathname
            for node in self.h5file.walk_nodes("/", classname="SoftLink")
        ]
        if common.verbose:
            print("detected links (classname='SoftLink'):", links)
        self.assertEqual(links, ["/larr2", "/lgroup1", "/group1/larr1"])

    def test08__v_links(self):
        """Checking `Group._v_links`."""

        links = [node for node in self.h5file.root._v_links]
        if common.verbose:
            print("detected links (under root):", links)
        self.assertEqual(len(links), 2)
        links = [node for node in self.h5file.root.group1._v_links]
        if common.verbose:
            print("detected links (under /group1):", links)
        self.assertEqual(links, ["larr1"])

    def test09_link_to_link(self):
        """Checking linked links."""

        # Create a link to another existing link
        lgroup2 = self.h5file.create_soft_link("/", "lgroup2", "/lgroup1")
        # Dereference it once:
        self.assertIs(lgroup2(), self.h5file.get_node("/lgroup1"))
        if common.verbose:
            print("First dereference is correct:", lgroup2())
        # Dereference it twice:
        self.assertIs(lgroup2()(), self.h5file.get_node("/group1"))
        if common.verbose:
            print("Second dereference is correct:", lgroup2()())

    def test10_copy_link_to_file(self):
        """Checking copying a link to another file."""

        fname = tempfile.mktemp(".h5")
        h5f = tb.open_file(fname, "a")
        h5f.create_array("/", "arr1", [1, 2])
        h5f.create_group("/", "group1")
        lgroup1 = self.h5file.root.lgroup1
        lgroup1_ = lgroup1.copy(h5f.root, "lgroup1")
        self.assertIn("/lgroup1", self.h5file)
        self.assertIn("/lgroup1", h5f)
        self.assertIn(lgroup1_, h5f)
        if common.verbose:
            print("Copied link:", lgroup1_, "in:", lgroup1_._v_file.filename)
        h5f.close()
        Path(fname).unlink()

    def test11_direct_attribute_access(self):
        """Check direct get/set attributes via link-->target.attribute"""

        larr1 = self.h5file.get_node("/lgroup1/larr1")
        arr1 = self.h5file.get_node("/arr1")
        # get
        self.assertEqual(larr1.shape, (2,))
        self.assertEqual(larr1[:], [1, 2])
        # set
        larr1[0] = -1
        self.assertEqual(arr1[:], [-1, 2])

    def test12_access_child_node_attributes(self):
        """Check get/set attributes via link-->target.child.attribute"""

        lgroup1 = self.h5file.get_node("/lgroup1")
        arr2 = self.h5file.get_node("/group1/arr2")
        # get child attribute
        self.assertEqual(lgroup1.arr2[:], [1, 2, 3])
        # set child attribute
        lgroup1.arr2[0] = -1
        self.assertEqual(arr2[:], [-1, 2, 3])

    def test13_direct_attribute_access_via_chained_softlinks(self):
        """Check get/set access via link2-->link1-->target.child.attribute"""

        self.h5file.get_node("/lgroup1")
        arr2 = self.h5file.get_node("/group1/arr2")
        # multiple chained links
        l_lgroup1 = self.h5file.create_soft_link("/", "l_lgroup1", "/lgroup1")
        # get child attribute
        self.assertEqual(l_lgroup1.arr2[:], [1, 2, 3])
        # set child attribute
        l_lgroup1.arr2[0] = -1
        self.assertEqual(arr2[:], [-1, 2, 3])

    def test14_child_of_softlink_to_group(self):
        """Create an array whose parent is a softlink to another group"""

        self.h5file.get_node("/group1")
        lgroup1 = self.h5file.get_node("/lgroup1")
        self.h5file.create_array(lgroup1, "new_arr", obj=[1, 2, 3])
        new_arr2 = self.h5file.get_node("/group1/new_arr")
        self.assertEqual(new_arr2[:], [1, 2, 3])

    def test_str(self):
        s = str(self.h5file)
        self.assertEqual(len(re.findall(r"\(SoftLink\)", s)), 3)
        self.assertEqual(len(re.findall(r"\(dangling\)", s)), 0)

    def test_str_with_dangling_link(self):
        self.h5file.root.group1.arr2.remove()
        s = str(self.h5file)
        self.assertEqual(len(re.findall(r"\(SoftLink\)", s)), 3)
        self.assertEqual(len(re.findall(r"\(dangling\)", s)), 1)


# Test for external links
@common.unittest.skipIf(
    tb.file._FILE_OPEN_POLICY == "strict", 'FILE_OPEN_POLICY = "strict"'
)
class ExternalLinkTestCase(common.TempFileMixin, common.PyTablesTestCase):

    def setUp(self):
        super().setUp()

        self.extfname = tempfile.mktemp(".h5")
        self.exth5file = tb.open_file(self.extfname, "w")
        self._createFile()

    def tearDown(self):
        """Remove ``extfname``."""

        extfname = self.extfname
        self.exth5file.close()
        super().tearDown()

        # open_files = tables.file._open_files
        # if self.extfname in open_files:
        #     #assert False
        #     for handler in open_files.get_handlers_by_name(self.extfname):
        #         handler.close()

        Path(extfname).unlink()  # comment this for debugging purposes only

    def _createFile(self):
        self.h5file.create_array("/", "arr1", [1, 2])
        group1 = self.h5file.create_group("/", "group1")
        self.h5file.create_array(group1, "arr2", [1, 2, 3])

        # The external file
        extarr1 = self.exth5file.create_array("/", "arr1", [1, 2])
        self.assertIsNotNone(extarr1)
        extgroup1 = self.exth5file.create_group("/", "group1")
        extarr2 = self.exth5file.create_array(extgroup1, "arr2", [1, 2, 3])

        # Create external links
        lgroup1 = self.h5file.create_external_link(
            "/", "lgroup1", "%s:/group1" % self.extfname
        )
        self.assertIsNotNone(lgroup1)
        larr1 = self.h5file.create_external_link(
            group1, "larr1", "%s:/arr1" % self.extfname
        )
        self.assertIsNotNone(larr1)
        larr2 = self.h5file.create_external_link("/", "larr2", extarr2)
        self.assertIsNotNone(larr2)

        # Re-open the external file in 'r'ead-only mode
        self.exth5file.close()
        self.exth5file = tb.open_file(self.extfname, "r")

    def test00_create(self):
        """Creating soft links."""

        self._checkEqualityGroup(
            self.exth5file.root.group1, self.h5file.root.lgroup1()
        )
        self._checkEqualityLeaf(
            self.exth5file.root.arr1, self.h5file.root.group1.larr1()
        )
        self._checkEqualityLeaf(
            self.h5file.root.lgroup1().arr2, self.h5file.root.larr2()
        )

    def test01_open(self):
        """Opening a file with soft links."""

        self._reopen()
        self._checkEqualityGroup(
            self.exth5file.root.group1, self.h5file.root.lgroup1()
        )
        self._checkEqualityLeaf(
            self.exth5file.root.arr1, self.h5file.root.group1.larr1()
        )
        self._checkEqualityLeaf(
            self.h5file.root.lgroup1().arr2, self.h5file.root.larr2()
        )

    def test02_remove(self):
        """Removing an external link."""

        # Re-open the external file in 'a'ppend mode
        self.exth5file.close()
        self.exth5file = tb.open_file(self.extfname, "a")

        # First delete the referred link
        self.exth5file.root.arr1.remove()
        self.assertNotIn("/arr1", self.exth5file)

        # The external link should still be there (but dangling)
        if common.verbose:
            print("Dangling link:", self.h5file.root.group1.larr1)
        self.assertIn("/group1/larr1", self.h5file)

        # Remove the external link itself
        self.h5file.root.group1.larr1.remove()
        self.assertNotIn("/group1/larr1", self.h5file)

    def test03_copy(self):
        """Copying an external link."""

        # Copy the link into another location
        root = self.h5file.root
        lgroup1 = root.lgroup1
        lgroup2 = lgroup1.copy("/", "lgroup2")
        self.assertIn("/lgroup1", self.h5file)
        self.assertIn("/lgroup2", self.h5file)
        self.assertIn("lgroup2", root._v_children)
        self.assertIn("lgroup2", root._v_links)
        if common.verbose:
            print("Copied link:", lgroup2)

        # Remove the first link
        lgroup1.remove()
        self._checkEqualityGroup(
            self.exth5file.root.group1, self.h5file.root.lgroup2()
        )

    def test03_overwrite(self):
        """Overwrite an external link."""

        # Copy the link into another location
        root = self.h5file.root
        lgroup1 = root.lgroup1
        lgroup2 = lgroup1.copy("/", "lgroup2")
        lgroup2 = lgroup1.copy("/", "lgroup2", overwrite=True)
        self.assertIn("/lgroup1", self.h5file)
        self.assertIn("/lgroup2", self.h5file)
        self.assertIn("lgroup2", root._v_children)
        self.assertIn("lgroup2", root._v_links)
        if common.verbose:
            print("Copied link:", lgroup2)

        # Remove the first link
        lgroup1.remove()
        self._checkEqualityGroup(
            self.exth5file.root.group1, self.h5file.root.lgroup2()
        )

    def test04_move(self):
        """Moving an external link."""

        # Move the link into another location
        lgroup1 = self.h5file.root.lgroup1
        group2 = self.h5file.create_group("/", "group2")
        lgroup1.move(group2, "lgroup2")
        lgroup2 = self.h5file.root.group2.lgroup2
        if common.verbose:
            print("Moved link:", lgroup2)
        self.assertNotIn("/lgroup1", self.h5file)
        self.assertIn("/group2/lgroup2", self.h5file)
        self._checkEqualityGroup(
            self.exth5file.root.group1, self.h5file.root.group2.lgroup2()
        )

    def test05_rename(self):
        """Renaming an external link."""

        # Rename the link
        lgroup1 = self.h5file.root.lgroup1
        lgroup1.rename("lgroup2")
        lgroup2 = self.h5file.root.lgroup2
        if common.verbose:
            print("Moved link:", lgroup2)
        self.assertNotIn("/lgroup1", self.h5file)
        self.assertIn("/lgroup2", self.h5file)
        self._checkEqualityGroup(
            self.exth5file.root.group1, self.h5file.root.lgroup2()
        )

    def test07_walkNodes(self):
        """Checking `walk_nodes` with `classname` option."""

        # Create a new soft link
        self.h5file.create_soft_link("/group1", "lgroup3", "./group3")
        links = [
            node._v_pathname
            for node in self.h5file.walk_nodes("/", classname="Link")
        ]
        if common.verbose:
            print("detected links (classname='Link'):", links)
        self.assertEqual(
            links, ["/larr2", "/lgroup1", "/group1/larr1", "/group1/lgroup3"]
        )
        links = [
            node._v_pathname
            for node in self.h5file.walk_nodes("/", classname="ExternalLink")
        ]
        if common.verbose:
            print("detected links (classname='ExternalLink'):", links)
        self.assertEqual(links, ["/larr2", "/lgroup1", "/group1/larr1"])

    def test08__v_links(self):
        """Checking `Group._v_links`."""

        links = [node for node in self.h5file.root._v_links]
        if common.verbose:
            print("detected links (under root):", links)
        self.assertEqual(len(links), 2)
        links = [node for node in self.h5file.root.group1._v_links]
        if common.verbose:
            print("detected links (under /group1):", links)
        self.assertEqual(links, ["larr1"])

    def test09_umount(self):
        """Checking `umount()` method."""

        link = self.h5file.root.lgroup1
        self.assertIsNone(link.extfile)

        # Dereference an external node (and hence, 'mount' a file)
        enode = link()
        self.assertIsNotNone(enode)
        self.assertIsNotNone(link.extfile)

        # Umount the link
        link.umount()
        self.assertIsNone(link.extfile)

    def test10_copy_link_to_file(self):
        """Checking copying a link to another file."""

        h5fname2 = tempfile.mktemp(".h5")
        try:
            with tb.open_file(h5fname2, "a") as h5file2:
                h5file2.create_array("/", "arr1", [1, 2])
                h5file2.create_group("/", "group1")
                lgroup1 = self.h5file.root.lgroup1
                lgroup1_ = lgroup1.copy(h5file2.root, "lgroup1")
                self.assertIn("/lgroup1", self.h5file)
                self.assertIn("/lgroup1", h5file2)
                self.assertIn(lgroup1_, h5file2)
                if common.verbose:
                    print(
                        "Copied link:",
                        lgroup1_,
                        "in:",
                        lgroup1_._v_file.filename,
                    )
        finally:
            if Path(h5fname2).is_file():
                Path(h5fname2).unlink()

    def test11_copy_entire_file_with_hardlink_option(self):
        """Checking copying the entire file (that contains external links)
        in a similar way ptrepack does (with hardlink kwargs activated)"""

        h5fname2 = tempfile.mktemp(".h5")
        try:
            with tb.open_file(h5fname2, "a") as h5file2:
                self.h5file.root._f_copy_children(
                    h5file2.root, recursive=True, use_hardlinks=True
                )
                self.assertIn("/lgroup1", h5file2)
        finally:
            if Path(h5fname2).is_file():
                Path(h5fname2).unlink()


def suite():
    """Return a test suite consisting of all the test cases in the module."""

    theSuite = common.unittest.TestSuite()
    niter = 1
    # common.heavy = 1  # uncomment this only for testing purposes

    for i in range(niter):
        theSuite.addTest(common.make_suite(HardLinkTestCase))
        theSuite.addTest(common.make_suite(SoftLinkTestCase))
        theSuite.addTest(common.make_suite(ExternalLinkTestCase))

    return theSuite


if __name__ == "__main__":
    import sys

    common.parse_argv(sys.argv)
    common.print_versions()
    common.unittest.main(defaultTest="suite")



import sys
from io import StringIO
from unittest.mock import patch

import tables.scripts.ptdump as ptdump
import tables.scripts.pttree as pttree
import tables.scripts.ptrepack as ptrepack
from tables.tests import common


class ptrepackTestCase(common.PyTablesTestCase):
    """Test ptrepack"""

    @patch.object(ptrepack, "copy_leaf")
    @patch.object(ptrepack.tb, "open_file")
    def test_paths_windows(self, mock_open_file, mock_copy_leaf):
        """Checking handling of windows filenames: test gh-616"""

        # this filename has a semi-colon to check for
        # regression of gh-616
        src_fn = "D:\\window~1\\path\\000\\infile"
        src_path = "/"
        dst_fn = "another\\path\\"
        dst_path = "/path/in/outfile"

        argv = ["ptrepack", src_fn + ":" + src_path, dst_fn + ":" + dst_path]
        with patch.object(sys, "argv", argv):
            ptrepack.main()

        args, kwargs = mock_open_file.call_args_list[0]
        self.assertEqual(args, (src_fn, "r"))

        args, kwargs = mock_copy_leaf.call_args_list[0]
        self.assertEqual(args, (src_fn, dst_fn, src_path, dst_path))


class ptdumpTestCase(common.PyTablesTestCase):
    """Test ptdump"""

    @patch.object(ptdump.tb, "open_file")
    @patch("sys.stdout", new_callable=StringIO)
    def test_paths_windows(self, _, mock_open_file):
        """Checking handling of windows filenames: test gh-616"""

        # this filename has a semi-colon to check for
        # regression of gh-616 (in ptdump)
        src_fn = "D:\\window~1\\path\\000\\ptdump"
        src_path = "/"

        argv = ["ptdump", src_fn + ":" + src_path]
        with patch.object(sys, "argv", argv):
            ptdump.main()

        args, kwargs = mock_open_file.call_args_list[0]
        self.assertEqual(args, (src_fn, "r"))


class pttreeTestCase(common.PyTablesTestCase):
    """Test ptdump"""

    @patch.object(pttree.tb, "open_file")
    @patch.object(pttree, "get_tree_str")
    @patch("sys.stdout", new_callable=StringIO)
    def test_paths_windows(self, _, mock_get_tree_str, mock_open_file):
        """Checking handling of windows filenames: test gh-616"""

        # this filename has a semi-colon to check for
        # regression of gh-616 (in pttree)
        src_fn = "D:\\window~1\\path\\000\\pttree"
        src_path = "/"

        argv = ["pttree", src_fn + ":" + src_path]
        with patch.object(sys, "argv", argv):
            pttree.main()

        args, kwargs = mock_open_file.call_args_list[0]
        self.assertEqual(args, (src_fn, "r"))


def suite():
    theSuite = common.unittest.TestSuite()

    theSuite.addTest(common.make_suite(ptrepackTestCase))
    theSuite.addTest(common.make_suite(ptdumpTestCase))
    theSuite.addTest(common.make_suite(pttreeTestCase))

    return theSuite


if __name__ == "__main__":
    common.parse_argv(sys.argv)
    common.print_versions()
    common.unittest.main(defaultTest="suite")



"""This utility lets you look into the data and metadata of your data files.

Pass the flag -h to this for help on usage.

"""

import argparse
import operator

import tables as tb

# default options
options = argparse.Namespace(
    rng=slice(None),
    showattrs=0,
    verbose=0,
    dump=0,
    colinfo=0,
    idxinfo=0,
)


def dump_leaf(leaf):
    """Dump an HDF5 leaf node."""
    if options.verbose:
        print(repr(leaf))
    else:
        print(str(leaf))
    if options.showattrs:
        print(f"  {leaf.attrs!r}")
    if options.dump and not isinstance(leaf, tb.unimplemented.UnImplemented):
        print("  Data dump:")
        # print((leaf.read(options.rng.start, options.rng.stop,
        #        options.rng.step))
        # This is better for large objects
        if options.rng.start is None:
            start = 0
        else:
            start = options.rng.start
        if options.rng.stop is None:
            if leaf.shape != ():
                stop = leaf.shape[0]
        else:
            stop = options.rng.stop
        if options.rng.step is None:
            step = 1
        else:
            step = options.rng.step
        if leaf.shape == ():
            print("[SCALAR] %s" % (leaf[()]))
        else:
            for i in range(start, stop, step):
                print(f"[{i}] {leaf[i]}")

    if isinstance(leaf, tb.table.Table) and options.colinfo:
        # Show info of columns
        for colname in leaf.colnames:
            print(repr(leaf.cols._f_col(colname)))

    if isinstance(leaf, tb.table.Table) and options.idxinfo:
        # Show info of indexes
        for colname in leaf.colnames:
            col = leaf.cols._f_col(colname)
            if isinstance(col, tb.table.Column) and col.index is not None:
                idx = col.index
                print(repr(idx))


def dump_group(pgroup, sort=False):
    """Dump an HDF5 group."""
    node_kinds = pgroup._v_file._node_kinds[1:]
    what = pgroup._f_walk_groups()
    if sort:
        what = sorted(what, key=operator.attrgetter("_v_pathname"))
    for group in what:
        print(str(group))
        if options.showattrs:
            print(f"  {group._v_attrs!r}")
        for kind in node_kinds:
            for node in group._f_list_nodes(kind):
                if options.verbose or options.dump:
                    dump_leaf(node)
                else:
                    print(str(node))


def _get_parser():
    parser = argparse.ArgumentParser(
        description=(
            "The ptdump utility allows you look into the contents of your "
            "PyTables files. It lets you see not only the data but also "
            "the metadata (that is, the *structure* and additional "
            "information in the form of *attributes*)."
        )
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="dump more metainformation on nodes",
    )
    parser.add_argument(
        "-d",
        "--dump",
        action="store_true",
        help="dump data information on leaves",
    )
    parser.add_argument(
        "-a",
        "--showattrs",
        action="store_true",
        help="show attributes in nodes (only useful when -v or -d are active)",
    )
    parser.add_argument(
        "-s",
        "--sort",
        action="store_true",
        help="sort output by node name",
    )
    parser.add_argument(
        "-c",
        "--colinfo",
        action="store_true",
        help="""show info of columns in tables (only useful when -v or -d
        are active)""",
    )
    parser.add_argument(
        "-i",
        "--idxinfo",
        action="store_true",
        help="""show info of indexed columns (only useful when -v or -d are
        active)""",
    )
    parser.add_argument(
        "-R",
        "--range",
        dest="rng",
        metavar="RANGE",
        help=(
            "select a RANGE of rows (in the form 'start,stop,step') "
            "during the copy of *all* the leaves. "
            "Default values are 'None, None, 1', which means a copy of all "
            "the rows."
        ),
    )
    parser.add_argument(
        "src",
        metavar="filename[:nodepath]",
        help="name of the HDF5 file to dump",
    )

    return parser


def main():
    """Implement the main CLI interface."""
    parser = _get_parser()

    args = parser.parse_args(namespace=options)

    # Get the options
    if isinstance(args.rng, str):
        try:
            options.rng = eval("slice(" + args.rng + ")")
        except Exception:
            parser.error("Error when getting the range parameter.")
        else:
            args.dump = 1

    # Catch the files passed as the last arguments
    src = args.src.rsplit(":", 1)
    if len(src) == 1:
        filename, nodename = src[0], "/"
    else:
        filename, nodename = src
        if nodename == "":
            # case where filename == "filename:" instead of "filename:/"
            nodename = "/"

    try:
        h5file = tb.open_file(filename, "r")
    except Exception as e:
        return f"Cannot open input file: {e}"

    with h5file:
        # Check whether the specified node is a group or a leaf
        nodeobject = h5file.get_node(nodename)
        if isinstance(nodeobject, tb.group.Group):
            # Close the file again and reopen using the root_uep
            dump_group(nodeobject, args.sort)
        elif isinstance(nodeobject, tb.leaf.Leaf):
            # If it is not a Group, it must be a Leaf
            dump_leaf(nodeobject)
        else:
            # This should never happen
            print("Unrecognized object:", nodeobject)



from time import perf_counter as clock
from pathlib import Path

import tables as tb

tref = clock()
trel = tref


def show_mem(explain):
    global tref, trel

    for line in Path("/proc/self/status").read_text().splitlines():
        if line.startswith("VmSize:"):
            vmsize = int(line.split()[1])
        elif line.startswith("VmRSS:"):
            vmrss = int(line.split()[1])
        elif line.startswith("VmData:"):
            vmdata = int(line.split()[1])
        elif line.startswith("VmStk:"):
            vmstk = int(line.split()[1])
        elif line.startswith("VmExe:"):
            vmexe = int(line.split()[1])
        elif line.startswith("VmLib:"):
            vmlib = int(line.split()[1])

    print("\nMemory usage: ******* %s *******" % explain)
    print(f"VmSize: {vmsize:>7} kB\tVmRSS: {vmrss:>7} kB")
    print(f"VmData: {vmdata:>7} kB\tVmStk: {vmstk:>7} kB")
    print(f"VmExe:  {vmexe:>7} kB\tVmLib: {vmlib:>7} kB")
    print("WallClock time:", clock() - tref, end=" ")
    print("  Delta time:", clock() - trel)
    trel = clock()


def write_group(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="w")
        for child in range(nchildren):
            fileh.create_group(
                fileh.root, "group" + str(child), "child: %d" % child
            )
        show_mem("After creating. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def read_group(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="r")
        for child in range(nchildren):
            node = fileh.get_node(fileh.root, "group" + str(child))
            assert node is not None
            # flavor = node._v_attrs.CLASS
        # for child in fileh.walk_nodes():
        #     pass
        show_mem("After reading metadata. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def write_array(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="w")
        for child in range(nchildren):
            fileh.create_array(
                fileh.root, "array" + str(child), [1, 1], "child: %d" % child
            )
        show_mem("After creating. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def read_array(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="r")
        for child in range(nchildren):
            node = fileh.get_node(fileh.root, "array" + str(child))
            # flavor = node._v_attrs.FLAVOR
            data = node[:]  # Read data
            assert data is not None
        show_mem("After reading data. Iter %s" % i)
        # for child in range(nchildren):
        #     node = fileh.get_node(fileh.root, 'array' + str(child))
        #     flavor = node._v_attrs.FLAVOR
        #     # flavor = node._v_attrs
        # for child in fileh.walk_nodes():
        #     pass
        # show_mem("After reading metadata. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def write_carray(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="w")
        for child in range(nchildren):
            fileh.create_carray(
                fileh.root,
                "array" + str(child),
                tb.IntAtom(),
                (2,),
                "child: %d" % child,
            )
        show_mem("After creating. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def read_carray(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="r")
        for child in range(nchildren):
            node = fileh.get_node(fileh.root, "array" + str(child))
            # flavor = node._v_attrs.FLAVOR
            data = node[:]  # Read data
            assert data is not None
            # print("data-->", data)
        show_mem("After reading data. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def write_earray(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="w")
        for child in range(nchildren):
            ea = fileh.create_earray(
                fileh.root,
                "array" + str(child),
                tb.IntAtom(),
                shape=(0,),
                title="child: %d" % child,
            )
            ea.append([1, 2, 3])
        show_mem("After creating. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def read_earray(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="r")
        for child in range(nchildren):
            node = fileh.get_node(fileh.root, "array" + str(child))
            # flavor = node._v_attrs.FLAVOR
            data = node[:]  # Read data
            assert data is not None
            # print("data-->", data)
        show_mem("After reading data. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def write_vlarray(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="w")
        for child in range(nchildren):
            vl = fileh.create_vlarray(
                fileh.root,
                "array" + str(child),
                tb.IntAtom(),
                "child: %d" % child,
            )
            vl.append([1, 2, 3])
        show_mem("After creating. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def read_vlarray(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="r")
        for child in range(nchildren):
            node = fileh.get_node(fileh.root, "array" + str(child))
            # flavor = node._v_attrs.FLAVOR
            data = node[:]  # Read data
            assert data is not None
            # print("data-->", data)
        show_mem("After reading data. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def write_table(filename, nchildren, niter):

    class Record(tb.IsDescription):
        var1 = tb.IntCol(pos=1)
        var2 = tb.StringCol(length=1, pos=2)
        var3 = tb.FloatCol(pos=3)

    for i in range(niter):
        fileh = tb.open_file(filename, mode="w")
        for child in range(nchildren):
            t = fileh.create_table(
                fileh.root, "table" + str(child), Record, "child: %d" % child
            )
            t.append([[1, "2", 3.0]])
        show_mem("After creating. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def read_table(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="r")
        for child in range(nchildren):
            node = fileh.get_node(fileh.root, "table" + str(child))
            # klass = node._v_attrs.CLASS
            data = node[:]  # Read data
            assert data is not None
            # print("data-->", data)
        show_mem("After reading data. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def write_xtable(filename, nchildren, niter):

    class Record(tb.IsDescription):
        var1 = tb.IntCol(pos=1)
        var2 = tb.StringCol(length=1, pos=2)
        var3 = tb.FloatCol(pos=3)

    for i in range(niter):
        fileh = tb.open_file(filename, mode="w")
        for child in range(nchildren):
            t = fileh.create_table(
                fileh.root, "table" + str(child), Record, "child: %d" % child
            )
            t.append([[1, "2", 3.0]])
            t.cols.var1.create_index()
        show_mem("After creating. Iter %s" % i)
        fileh.close()
        show_mem("After close")


def read_xtable(filename, nchildren, niter):
    for i in range(niter):
        fileh = tb.open_file(filename, mode="r")
        for child in range(nchildren):
            node = fileh.get_node(fileh.root, "table" + str(child))
            # klass = node._v_attrs.CLASS
            # data = node[:]  # Read data
            # print("data-->", data)
        show_mem("After reading data. Iter %s" % i)
        fileh.close()
        show_mem("After close")
        del node


if __name__ == "__main__":
    import pstats
    import profile as prof
    import argparse

    def _get_parser():
        parser = argparse.ArgumentParser(
            description="Check for PyTables memory leaks."
        )
        parser.add_argument(
            "-v", "--verbose", action="store_true", help="enable verbose mode"
        )
        parser.add_argument(
            "-p", "--profile", action="store_true", help="profile"
        )
        parser.add_argument(
            "-a",
            "--array",
            action="store_true",
            help="create/read arrays (default)",
        )
        parser.add_argument(
            "-c", "--carray", action="store_true", help="create/read carrays"
        )
        parser.add_argument(
            "-e", "--earray", action="store_true", help="create/read earrays"
        )
        parser.add_argument(
            "-l", "--vlarray", action="store_true", help="create/read vlarrays"
        )
        parser.add_argument(
            "-t", "--table", action="store_true", help="create/read tables"
        )
        parser.add_argument(
            "-x",
            "--indexed-table",
            action="store_true",
            dest="xtable",
            help="create/read indexed-tables",
        )
        parser.add_argument(
            "-g", "--group", action="store_true", help="create/read groups"
        )
        parser.add_argument(
            "-r", "--read", action="store_true", help="only read test"
        )
        parser.add_argument(
            "-w", "--write", action="store_true", help="only write test"
        )
        parser.add_argument(
            "-n",
            "--nchildren",
            type=int,
            default=1000,
            help="number of children (%(default)d is the " "default)",
        )
        parser.add_argument(
            "-i",
            "--niter",
            type=int,
            default=3,
            help="number of iterations (default: %(default)d)",
        )

        parser.add_argument("filename", help="HDF5 file name")

        return parser

    parser = _get_parser()
    args = parser.parse_args()

    # set 'array' as default value if no ather option has been specified
    for name in ("carray", "earray", "vlarray", "table", "xtable", "group"):
        if getattr(args, name):
            break
    else:
        args.array = True

    filename = args.filename
    nchildren = args.nchildren
    niter = args.niter

    if args.array:
        fwrite = "write_array"
        fread = "read_array"
    elif args.carray:
        fwrite = "write_carray"
        fread = "read_carray"
    elif args.earray:
        fwrite = "write_earray"
        fread = "read_earray"
    elif args.vlarray:
        fwrite = "write_vlarray"
        fread = "read_vlarray"
    elif args.table:
        fwrite = "write_table"
        fread = "read_table"
    elif args.xtable:
        fwrite = "write_xtable"
        fread = "read_xtable"
    elif args.group:
        fwrite = "write_group"
        fread = "read_group"

    show_mem("Before open")
    if args.write:
        if args.profile:
            prof.run(
                str(fwrite) + "(filename, nchildren, niter)", "write_file.prof"
            )
            stats = pstats.Stats("write_file.prof")
            stats.strip_dirs()
            stats.sort_stats("time", "calls")
            if args.verbose:
                stats.print_stats()
            else:
                stats.print_stats(20)
        else:
            eval(fwrite + "(filename, nchildren, niter)")
    if args.read:
        if args.profile:
            prof.run(fread + "(filename, nchildren, niter)", "read_file.prof")
            stats = pstats.Stats("read_file.prof")
            stats.strip_dirs()
            stats.sort_stats("time", "calls")
            if args.verbose:
                print("profile -verbose")
                stats.print_stats()
            else:
                stats.print_stats(20)
        else:
            eval(fread + "(filename, nchildren, niter)")



"""Run unit tests in parallel threads."""

# Based on unittest-ft, which is copyright Amethyst Reese, MIT license.

from __future__ import annotations

import os
import sys
import time
import random
import logging
import argparse
import unittest
import threading
import collections
from typing import Any, Self, TextIO
from dataclasses import dataclass, field
from collections.abc import Generator
from concurrent.futures import (
    Future,
    ThreadPoolExecutor,
    as_completed,
)

LOG = logging.getLogger(__name__)

DEFAULT_THREADS = os.cpu_count() or 4


class FTTestResult(unittest.TestResult):
    def __init__(
        self,
        stream: TextIO | None = None,
        descriptions: bool | None = None,
        verbosity: int | None = None,
    ) -> None:
        super().__init__(
            stream=stream, descriptions=descriptions, verbosity=verbosity
        )
        self.verbosity = verbosity or 1
        self.before = time.monotonic_ns()
        self.duration = 0
        self.collected_duration = 0

    def stopTest(self, test: Any) -> None:
        super().stopTest(test)
        self.duration = time.monotonic_ns() - self.before

    def stopTestRun(self) -> None:
        super().stopTestRun()
        self.duration = time.monotonic_ns() - self.before

    def __str__(self) -> str:
        items: dict[tuple[str, str], int] = collections.defaultdict(int)
        for test_case, trace in self.errors:
            items[f"ERROR: {test_case}", trace] += 1
        for test_case, trace in self.failures:
            items[f"FAIL: {test_case}", trace] += 1

        results = {
            f"{label}": trace for (label, trace), count in items.items()
        }
        longest = max(len(label) for label in results) if results else 70

        msg = "\n"
        msg += "\n".join(
            f"{'=' * longest}\n{label}\n{'-' * longest}\n{trace}"
            for label, trace in results.items()
        )
        msg += "-" * longest
        msg += f"\nRan {self.testsRun} tests in {format_ns(self.duration)}"

        saved = self.collected_duration - self.duration
        if saved > 0 and (saved / self.duration) > 0.10:
            msg += f" (saved {format_ns(self.collected_duration - self.duration)})"
        msg += "\n\n"

        msg += "OK" if self.wasSuccessful() else "FAILED"

        parts = []
        if self.errors:
            parts += [f"errors={len(self.errors)}"]
        if self.failures:
            parts += [f"failures={len(self.failures)}"]
        if self.skipped:
            parts += [f"skipped={len(self.skipped)}"]
        if self.expectedFailures:
            parts += [f"expected failures={len(self.expectedFailures)}"]

        if parts:
            msg += f" ({', '.join(parts)})"

        return msg

    def __add__(self, other: object) -> FTTestResult:
        if not isinstance(other, unittest.TestResult):
            return NotImplemented
        result = FTTestResult()
        result.errors = self.errors + other.errors
        result.expectedFailures = (
            self.expectedFailures + other.expectedFailures
        )
        result.failures = self.failures + other.failures
        result.skipped = self.skipped + other.skipped
        result.testsRun = self.testsRun + other.testsRun
        result.unexpectedSuccesses = (
            self.unexpectedSuccesses + other.unexpectedSuccesses
        )
        if isinstance(other, FTTestResult):
            result.collected_duration = self.duration + other.duration
        return result

    def __iadd__(self, other: object) -> Self:
        if not isinstance(other, unittest.TestResult):
            return NotImplemented
        self.errors += other.errors
        self.expectedFailures += other.expectedFailures
        self.failures += other.failures
        self.skipped += other.skipped
        self.testsRun += other.testsRun
        self.unexpectedSuccesses += other.unexpectedSuccesses
        if isinstance(other, FTTestResult):
            self.collected_duration += other.duration
        return self


def get_individual_tests(
    suite: unittest.TestSuite,
) -> Generator[unittest.TestCase]:
    for test in suite:
        if isinstance(test, unittest.TestSuite):
            yield from get_individual_tests(test)
        else:
            yield test


def run_single_test(suite: unittest.TestSuite) -> tuple[str, FTTestResult]:
    test_id = suite.id()
    LOG.debug(f"Running test {threading.get_ident()} {test_id}")
    result = FTTestResult(descriptions=True, verbosity=2)
    suite.run(result)
    LOG.debug("Finished test %s", test_id)
    return (test_id, result)


def format_ns(duration: int) -> str:
    if duration < 1_000_000_000:
        return f"{duration / 1_000_000:.2f}ms"
    else:
        return f"{duration / 1_000_000_000:.3f}s"


@dataclass
class Output:
    total: int
    futures: dict[Future[tuple[str, FTTestResult]], str] = field(
        default_factory=dict
    )
    stream: TextIO = sys.stdout
    verbosity: int = 1

    def __post_init__(self) -> None:
        self.count = 0

    def render(self, test_id: str, test_result: FTTestResult) -> None:
        stream = self.stream
        verbosity = self.verbosity

        self.count += 1
        if verbosity == 2:
            stream.write(
                f"[{self.count}/{self.total}] {test_id}"
                f" ... {'OK' if test_result.wasSuccessful() else 'FAIL'} "
                f" {format_ns(test_result.duration)}\n"
            )
        elif verbosity == 1:
            if test_result.errors:
                stream.write("E")
            elif test_result.failures:
                stream.write("F")
            elif test_result.expectedFailures:
                stream.write("x")
            elif test_result.skipped:
                stream.write("s")
            else:
                stream.write(".")
        stream.flush()


_EXCLUDE_CASES = set("""
    tables.filters.Filters
    tables.misc.enum
    tables.tests.test_array.SI1NACloseTestCase
    tables.tests.test_array.SI1NAOpenTestCase
    tables.tests.test_array.SI2NACloseTestCase
    tables.tests.test_array.SI2NAOpenTestCase
    tables.tests.test_basics.HDF5ErrorHandling
    tables.tests.test_basics.OpenFileFailureTestCase
    tables.tests.test_create.SetBloscMaxThreadsTestCase
    tables.tests.test_tablesMD.CompressTwoTablesTestCase
    tables.tests.test_utils.ptdumpTestCase.test_paths_windows
    tables.tests.test_utils.ptrepackTestCase.test_paths_windows
    tables.tests.test_utils.pttreeTestCase.test_paths_windows
    """.strip().split())


def _match_case(test_id: str) -> bool:
    for pat in _EXCLUDE_CASES:
        if test_id.startswith(pat):
            return True
    return False


def run_suite(
    suite: unittest.TestSuite,
    *,
    randomize: bool = False,
    threads: int = DEFAULT_THREADS,
    verbosity: int = 1,
    max_tests: int = 0,
) -> unittest.TestResult:
    test_cases = {}
    for test_case in get_individual_tests(suite):
        test_id = test_case.id()
        if _match_case(test_id):
            continue
        test_cases[test_id] = test_case
    test_ids = list(test_cases)
    if randomize:
        rnd = random.SystemRandom()
        rnd.shuffle(test_ids)
    else:
        test_ids.sort()
    if max_tests:
        test_ids = test_ids[:max_tests]

    LOG.info(
        "Ready to run %d tests:\n  %s", len(test_ids), "\n  ".join(test_ids)
    )

    output = Output(total=len(test_ids), verbosity=verbosity)
    result = FTTestResult()

    with ThreadPoolExecutor(max_workers=threads) as pool:
        futures = [
            pool.submit(run_single_test, test_cases[test_id])
            for test_id in test_ids
        ]
        for fut in as_completed(futures):
            test_id, test_result = fut.result()
            result += test_result
            output.render(test_id, test_result)
        result.stopTestRun()

    return result


def main():
    parser = argparse.ArgumentParser(
        description="Run PyTables tests with free-threading support"
    )
    parser.add_argument(
        "--randomize",
        action="store_true",
        help="Run tests in random order",
    )
    parser.add_argument(
        "--verbosity",
        type=int,
        default=1,
        help="Logger verbosity level (default: 1)",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=DEFAULT_THREADS,
        help=f"Number of worker threads to use (default: {DEFAULT_THREADS})",
    )
    parser.add_argument(
        "--max-tests",
        type=int,
        default=0,
        help="Maximum number of tests to run, 0 = unlimited (default: 0)",
    )
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.verbosity > 1 else logging.INFO
    )

    from tables.tests import test_suite

    suite = test_suite.suite()
    result = run_suite(
        suite,
        verbosity=args.verbosity,
        threads=args.threads,
        randomize=args.randomize,
        max_tests=args.max_tests,
    )
    print(result)


if __name__ == "__main__":
    main()



import unittest

import numpy as np

import tables as tb
from tables.tests import common


class TestAuxiliaryFunctions(unittest.TestCase):
    def test_keysort(self):
        N = 1000
        rnd = np.random.randint(N, size=N)
        for dtype1 in ("S6", "b1", "i1", "i8", "u4", "u8", "f4", "f8"):
            for dtype2 in ("u4", "i8"):
                a = np.array(rnd, dtype1)
                b = np.array(rnd, dtype2)

                c = a.copy()
                d = c.argsort()
                e = c[d]
                f = b[d]

                tb.indexesextension.keysort(a, b)
                self.assertTrue((a == e).all())
                self.assertTrue((b == f).all())


def suite():
    theSuite = unittest.TestSuite()
    theSuite.addTest(common.make_suite(TestAuxiliaryFunctions))
    return theSuite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



import sys

import numpy as np
from packaging.version import parse as parse_version

import tables as tb
from tables.tests import common


# Test Record class
class Record(tb.IsDescription):
    var1 = tb.StringCol(itemsize=4)  # 4-character String
    var2 = tb.Col.from_kind("int")  # integer
    var3 = tb.Col.from_kind("int", itemsize=2)  # short integer
    var4 = tb.Col.from_kind("float")  # double (double-precision)
    var5 = tb.Col.from_kind("float", itemsize=4)  # float  (single-precision)
    var6 = tb.Col.from_kind("complex")  # double-precision
    var7 = tb.Col.from_kind("complex", itemsize=8)  # single-precision
    if hasattr(tb, "Float16Atom"):
        var8 = tb.Col.from_kind("float", itemsize=2)  # half-precision
    if hasattr(tb, "Float96Atom"):
        var9 = tb.Col.from_kind("float", itemsize=12)  # extended-precision
    if hasattr(tb, "Float128Atom"):
        var10 = tb.Col.from_kind("float", itemsize=16)  # extended-precision
    if hasattr(tb, "Complex192Atom"):
        var11 = tb.Col.from_kind("complex", itemsize=24)  # extended-precision
    if hasattr(tb, "Complex256Atom"):
        var12 = tb.Col.from_kind("complex", itemsize=32)  # extended-precision


class RangeTestCase(common.TempFileMixin, common.PyTablesTestCase):
    title = "This is the table title"
    expectedrows = 100
    maxshort = 2**15
    maxint = 2_147_483_648  # (2 ** 31)
    compress = 0

    def setUp(self):
        super().setUp()
        self.rootgroup = self.h5file.root

        # Create a table
        self.table = self.h5file.create_table(
            self.rootgroup, "table", Record, self.title
        )

    def test00_range(self):
        """Testing the range check."""

        rec = self.table.row

        # Save a record
        i = self.maxshort
        rec["var1"] = "%04d" % (i)
        rec["var2"] = i
        rec["var3"] = np.array(i).astype("i2")
        rec["var4"] = float(i)
        rec["var5"] = float(i)
        rec["var6"] = float(i)
        rec["var7"] = complex(i, i)
        if hasattr(tb, "Float16Atom"):
            rec["var8"] = float(i)
        if hasattr(tb, "Float96Atom"):
            rec["var9"] = float(i)
        if hasattr(tb, "Float128Atom"):
            rec["var10"] = float(i)
        try:
            rec.append()
        except ValueError:
            if common.verbose:
                type, value, traceback = sys.exc_info()
                print("\nGreat!, the next ValueError was catched!")
                print(value)
            pass
        else:
            if common.verbose:
                print(
                    "\nNow, the range overflow no longer issues a ValueError"
                )

    def test01_type(self):
        """Testing the type check."""

        rec = self.table.row
        # Save a record
        i = self.maxshort
        rec["var1"] = "%04d" % (i)
        rec["var2"] = i
        rec["var3"] = np.array(i % self.maxshort).astype("i2")
        rec["var5"] = float(i)

        # Numpy 1.25 -> ValueError
        with self.assertRaises((TypeError, ValueError)):
            rec["var4"] = "124c"

        rec["var6"] = float(i)
        rec["var7"] = complex(i, i)
        if hasattr(tb, "Float16Atom"):
            rec["var8"] = float(i)
        if hasattr(tb, "Float96Atom"):
            rec["var9"] = float(i)
        if hasattr(tb, "Float128Atom"):
            rec["var10"] = float(i)


# Check the dtype read-only attribute
class DtypeTestCase(common.TempFileMixin, common.PyTablesTestCase):

    def test00a_table(self):
        """Check dtype accessor for Table objects."""

        a = self.h5file.create_table("/", "table", Record)
        self.assertEqual(a.dtype, a.description._v_dtype)

    def test00b_column(self):
        """Check dtype accessor for Column objects."""

        a = self.h5file.create_table("/", "table", Record)
        c = a.cols.var3
        self.assertEqual(c.dtype, a.description._v_dtype["var3"])

    def test01_array(self):
        """Check dtype accessor for Array objects."""

        a = self.h5file.create_array("/", "array", [1, 2])
        self.assertEqual(a.dtype, a.atom.dtype)

    def test02_carray(self):
        """Check dtype accessor for CArray objects."""

        a = self.h5file.create_carray(
            "/", "array", atom=tb.FloatAtom(), shape=[1, 2]
        )
        self.assertEqual(a.dtype, a.atom.dtype)

    def test03_carray(self):
        """Check dtype accessor for EArray objects."""

        a = self.h5file.create_earray(
            "/", "array", atom=tb.FloatAtom(), shape=[0, 2]
        )
        self.assertEqual(a.dtype, a.atom.dtype)

    def test04_vlarray(self):
        """Check dtype accessor for VLArray objects."""

        a = self.h5file.create_vlarray("/", "array", tb.FloatAtom())
        self.assertEqual(a.dtype, a.atom.dtype)


class ReadFloatTestCase(common.TestFileMixin, common.PyTablesTestCase):
    h5fname = common.test_filename("float.h5")
    nrows = 5
    ncols = 6

    def setUp(self):
        super().setUp()
        x = np.arange(self.ncols)
        y = np.arange(self.nrows)
        y.shape = (self.nrows, 1)
        self.values = x + y

    def test01_read_float16(self):
        dtype = "float16"
        if hasattr(np, dtype):
            ds = getattr(self.h5file.root, dtype)
            self.assertNotIsInstance(ds, tb.UnImplemented)
            self.assertEqual(ds.shape, (self.nrows, self.ncols))
            self.assertEqual(ds.dtype, dtype)
            self.assertTrue(
                common.allequal(ds.read(), self.values.astype(dtype))
            )
        else:
            with self.assertWarns(UserWarning):
                ds = getattr(self.h5file.root, dtype)
            self.assertIsInstance(ds, tb.UnImplemented)

    def test02_read_float32(self):
        dtype = "float32"
        ds = getattr(self.h5file.root, dtype)
        self.assertNotIsInstance(ds, tb.UnImplemented)
        self.assertEqual(ds.shape, (self.nrows, self.ncols))
        self.assertEqual(ds.dtype, dtype)
        self.assertTrue(common.allequal(ds.read(), self.values.astype(dtype)))

    def test03_read_float64(self):
        dtype = "float64"
        ds = getattr(self.h5file.root, dtype)
        self.assertNotIsInstance(ds, tb.UnImplemented)
        self.assertEqual(ds.shape, (self.nrows, self.ncols))
        self.assertEqual(ds.dtype, dtype)
        self.assertTrue(common.allequal(ds.read(), self.values.astype(dtype)))

    def test04_read_longdouble(self):
        dtype = "longdouble"
        if hasattr(tb, "Float96Atom") or hasattr(tb, "Float128Atom"):
            ds = getattr(self.h5file.root, dtype)
            self.assertNotIsInstance(ds, tb.UnImplemented)
            self.assertEqual(ds.shape, (self.nrows, self.ncols))
            self.assertEqual(ds.dtype, dtype)
            self.assertTrue(
                common.allequal(ds.read(), self.values.astype(dtype))
            )

            if hasattr(tb, "Float96Atom"):
                self.assertEqual(ds.dtype, "float96")
            elif hasattr(tb, "Float128Atom"):
                self.assertEqual(ds.dtype, "float128")
        else:
            # XXX: check
            # the behavior depends on the HDF5 lib configuration
            try:
                with self.assertWarns(UserWarning):
                    ds = getattr(self.h5file.root, dtype)
                self.assertIsInstance(ds, tb.UnImplemented)
            except AssertionError:
                ds = getattr(self.h5file.root, dtype)
                self.assertEqual(ds.dtype, "float64")

    def test05_read_quadprecision_float(self):
        # XXX: check
        try:
            with self.assertWarns(UserWarning):
                ds = self.h5file.root.quadprecision
            self.assertIsInstance(ds, tb.UnImplemented)
        except AssertionError:
            # NOTE: it would be nice to have some sort of message that warns
            #       against the potential precision loss: the quad-precision
            #       dataset actually uses 128 bits for each element, not just
            #       80 bits (longdouble)
            ds = self.h5file.root.quadprecision
            self.assertEqual(ds.dtype, "longdouble")


class AtomTestCase(common.PyTablesTestCase):
    def test_init_parameters_01(self):
        atom1 = tb.StringAtom(itemsize=12)
        atom2 = atom1.copy()
        self.assertEqual(atom1, atom2)
        self.assertEqual(str(atom1), str(atom2))
        self.assertIsNot(atom1, atom2)

    def test_init_parameters_02(self):
        atom1 = tb.StringAtom(itemsize=12)
        atom2 = atom1.copy(itemsize=100, shape=(2, 2))
        self.assertEqual(
            atom2, tb.StringAtom(itemsize=100, shape=(2, 2), dflt=b"")
        )

    def test_init_parameters_03(self):
        atom1 = tb.StringAtom(itemsize=12)
        self.assertRaises(TypeError, atom1.copy, foobar=42)

    def test_from_dtype_01(self):
        atom1 = tb.Atom.from_dtype(np.dtype((np.int16, (2, 2))))
        atom2 = tb.Int16Atom(shape=(2, 2), dflt=0)
        self.assertEqual(atom1, atom2)
        self.assertEqual(str(atom1), str(atom2))

    def test_from_dtype_02(self):
        atom1 = tb.Atom.from_dtype(np.dtype("S5"), dflt=b"hello")
        atom2 = tb.StringAtom(itemsize=5, shape=(), dflt=b"hello")
        self.assertEqual(atom1, atom2)
        self.assertEqual(str(atom1), str(atom2))

    def test_from_dtype_03(self):
        with self.assertWarns(Warning):
            atom1 = tb.Atom.from_dtype(np.dtype("U5"), dflt=b"hello")
        atom2 = tb.StringAtom(itemsize=5, shape=(), dflt=b"hello")
        self.assertEqual(atom1, atom2)
        self.assertEqual(str(atom1), str(atom2))

    def test_from_dtype_04(self):
        atom1 = tb.Atom.from_dtype(np.dtype("float64"))
        atom2 = tb.Float64Atom(shape=(), dflt=0.0)
        self.assertEqual(atom1, atom2)
        self.assertEqual(str(atom1), str(atom2))

    def test_from_kind_01(self):
        atom1 = tb.Atom.from_kind("int", itemsize=2, shape=(2, 2))
        atom2 = tb.Int16Atom(shape=(2, 2), dflt=0)
        self.assertEqual(atom1, atom2)
        self.assertEqual(str(atom1), str(atom2))

    def test_from_kind_02(self):
        atom1 = tb.Atom.from_kind("int", shape=(2, 2))
        atom2 = tb.Int32Atom(shape=(2, 2), dflt=0)
        self.assertEqual(atom1, atom2)
        self.assertEqual(str(atom1), str(atom2))

    def test_from_kind_03(self):
        atom1 = tb.Atom.from_kind("int", shape=1)
        atom2 = tb.Int32Atom(shape=(1,), dflt=0)
        self.assertEqual(atom1, atom2)
        self.assertEqual(str(atom1), str(atom2))

    def test_from_kind_04(self):
        atom1 = tb.Atom.from_kind("string", itemsize=5, dflt=b"hello")
        atom2 = tb.StringAtom(itemsize=5, shape=(), dflt=b"hello")
        self.assertEqual(atom1, atom2)
        self.assertEqual(str(atom1), str(atom2))

    def test_from_kind_05(self):
        # ValueError: no default item size for kind ``string``
        self.assertRaises(
            ValueError, tb.Atom.from_kind, "string", dflt=b"hello"
        )

    def test_from_kind_06(self):
        # ValueError: unknown kind: 'Float'
        self.assertRaises(ValueError, tb.Atom.from_kind, "Float")


def suite():
    import doctest

    theSuite = common.unittest.TestSuite()

    for i in range(1):
        # TODO: in numpy 2 the repr of various dtypes has changed breaking the
        # doctests. When only numpy 2 is supported re-enable these tests.
        if parse_version(np.__version__) < parse_version("2.dev0"):
            theSuite.addTest(doctest.DocTestSuite(tb.atom))
        theSuite.addTest(common.make_suite(AtomTestCase))
        theSuite.addTest(common.make_suite(RangeTestCase))
        theSuite.addTest(common.make_suite(DtypeTestCase))
        theSuite.addTest(common.make_suite(ReadFloatTestCase))

    return theSuite


if __name__ == "__main__":
    common.parse_argv(sys.argv)
    common.print_versions()
    common.unittest.main(defaultTest="suite")
